# -*- coding: UTF-8 -*-
# pylint: disable=redefined-builtin,unused-import
"""
Provides some Python3 exception classes for Python2 and early Python3 versions.
"""

from __future__ import absolute_import
import errno as _errno
from six.moves import builtins as _builtins


# -----------------------------------------------------------------------------
# EXCEPTION CLASSES:
# -----------------------------------------------------------------------------
FileNotFoundError = getattr(_builtins, "FileNotFoundError", None)
if not FileNotFoundError:
    class FileNotFoundError(OSError):
        """Provided since Python >= 3.3"""
        errno = _errno.ENOENT


ModuleNotFoundError = getattr(_builtins, "ModuleNotFoundError", None)
if not ModuleNotFoundError:
    class ModuleNotFoundError(ImportError):
        """Provided since Python >= 3.6"""



# -*- coding: utf-8 -*-
"""
Importer module for lazy-loading/importing modules and objects.

REQUIRES: importlib (provided in Python2.7, Python3.2...)
"""

from __future__ import absolute_import
import importlib
import inspect
from behave._types import Unknown
from behave.exception import ClassNotFoundError, ModuleNotFoundError


def parse_scoped_name(scoped_name):
    """
    SCHEMA: my.module_name:MyClassName
    EXAMPLE:
        behave.formatter.plain:PlainFormatter
    """
    scoped_name = scoped_name.strip()
    if "::" in scoped_name:
        # -- ALTERNATIVE: my.module_name::MyClassName
        scoped_name = scoped_name.replace("::", ":")
    if ":" not in scoped_name:
        schema = "%s: Missing ':' (colon) as module-to-name seperator'"
        raise ValueError(schema % scoped_name)
    module_name, object_name = scoped_name.rsplit(":", 1)
    return module_name, object_name or ""


def make_scoped_class_name(obj):
    """Build scoped-class-name from an object/class.

    :param obj:  Object or class.
    :return Scoped-class-name (as string).
    """
    if inspect.isclass(obj):
        class_name = obj.__name__
    else:
        class_name = obj.__class__.__name__
    module_name = getattr(obj, "__module__", None)
    if module_name:
        return "{0}:{1}".format(obj.__module__, class_name)
    # -- OTHERWISE: Builtin data type
    return class_name


def load_module(module_name):
    try:
        return importlib.import_module(module_name)
    except ModuleNotFoundError:
        # -- SINCE: Python 3.6 (special kind of ImportError)
        raise
    except ImportError as e:
        # -- CASE: Python < 3.6 (Python 2.7, ...)
        msg = str(e)
        if not msg.endswith("'"):
            # -- NOTE: Emulate ModuleNotFoundError message:
            #    "No module named '{module_name}'"
            prefix, module_name = msg.rsplit(" ", 1)
            msg = "{0} '{1}'".format(prefix, module_name)
        raise ModuleNotFoundError(msg)


class LazyObject(object):
    """Provides a placeholder for an class/object that should be loaded lazily.

    It stores the module-name, object-name/class-name and
    imports it later (on demand) when this lazy-object is accessed.
    """

    def __init__(self, module_name, object_name=None):
        if ":" in module_name and not object_name:
            module_name, object_name = parse_scoped_name(module_name)
        assert ":" not in module_name
        self.module_name = module_name
        self.object_name = object_name
        self.resolved_object = None
        self.error = None

    # -- PYTHON DESCRIPTOR PROTOCOL:
    def __get__(self, obj=None, type=None):     # pylint: disable=redefined-builtin
        """Implement descriptor protocol,
        useful if this class is used as attribute.

        :return: Real object (lazy-loaded if necessary).
        :raise ModuleNotFoundError: If module is not found or cannot be imported.
        :raise ClassNotFoundError:  If class/object is not found in module.
        """
        __pychecker__ = "unusednames=obj,type"
        resolved_object = None
        if not self.resolved_object:
            # -- SETUP-ONCE: Lazy load the real object.
            try:
                module = load_module(self.module_name)
                resolved_object = getattr(module, self.object_name, Unknown)
                if resolved_object is Unknown:
                    # OLD: msg = "%s: %s is Unknown" % (self.module_name, self.object_name)
                    scoped_name = "%s:%s" % (self.module_name, self.object_name)
                    raise ClassNotFoundError(scoped_name)
                self.resolved_object = resolved_object
            except ImportError as e:
                self.error = "%s: %s" % (e.__class__.__name__, e)
                raise
                # OR: resolved_object = self
        return resolved_object

    def __set__(self, obj, value):
        """Implement descriptor protocol."""
        __pychecker__ = "unusednames=obj"
        self.resolved_object = value

    def get(self):
        return self.__get__()


class LazyDict(dict):
    """Provides a dict that supports lazy loading of classes/objects.
    A LazyObject is provided as placeholder for a value that should be
    loaded lazily.

    EXAMPLE:

    .. code-block:: python

        from behave.importer import LazyDict

        the_plugin_registry = LazyDict({
            "alice": LazyObject("my_module.alice_plugin:AliceClass"),
            "bob": LayzObject("my_module.bob_plugin:BobClass"),
        })

        # -- LATER: Import plugin-class module(s) only if needed.
        # INTENTION: Pay only (with runtime costs) for what you use.
        config.plugin_name = "alice"
        plugin_class = the_plugin_registry[config.plugin_name]
        ...
    """

    def __getitem__(self, key):
        """Provides access to the stored dict value(s).

        Implements lazy loading of item value (if necessary).
        When lazy object is loaded, its value with the dict is replaced
        with the real value.

        :param key:  Key to access the value of an item in the dict.
        :return: value
        :raises KeyError: if item is not found.
        :raises ModuleNotFoundError: for a LazyObject module is not found.
        :raises ClassNotFoundError:  for a LazyObject class/object is not found in module.
        """
        value = dict.__getitem__(self, key)
        if isinstance(value, LazyObject):
            # -- LAZY-LOADING MECHANISM:
            # Load class/object once and replace the lazy placeholder.
            value = value.__get__()
            self[key] = value
        return value

    def load_all(self, strict=False):
        for key in self.keys():
            try:
                self.__getitem__(key)
            except ImportError:
                if strict:
                    raise



# -*- coding: UTF-8 -*-
"""
Provides a step registry and step decorators.
The step registry allows to match steps (model elements) with
step implementations (step definitions). This is necessary to execute steps.
"""

from __future__ import absolute_import
from behave.matchers import Match, get_matcher
from behave.textutil import text as _text

# limit import * to just the decorators
# pylint: disable=undefined-all-variable
# names = "given when then step"
# names = names + " " + names.title()
# __all__ = names.split()
__all__ = [
    "given", "when", "then", "step",    # PREFERRED.
    "Given", "When", "Then", "Step"     # Also possible.
]


class AmbiguousStep(ValueError):
    pass


class StepRegistry(object):
    def __init__(self):
        self.steps = {
            "given": [],
            "when": [],
            "then": [],
            "step": [],
        }

    @staticmethod
    def same_step_definition(step, other_pattern, other_location):
        return (step.pattern == other_pattern and
                step.location == other_location and
                other_location.filename != "<string>")

    def add_step_definition(self, keyword, step_text, func):
        step_location = Match.make_location(func)
        step_type = keyword.lower()
        step_text = _text(step_text)
        step_definitions = self.steps[step_type]
        for existing in step_definitions:
            if self.same_step_definition(existing, step_text, step_location):
                # -- EXACT-STEP: Same step function is already registered.
                # This may occur when a step module imports another one.
                return
            elif existing.match(step_text):     # -- SIMPLISTIC
                message = u"%s has already been defined in\n  existing step %s"
                new_step = u"@%s('%s')" % (step_type, step_text)
                existing.step_type = step_type
                existing_step = existing.describe()
                existing_step += u" at %s" % existing.location
                raise AmbiguousStep(message % (new_step, existing_step))
        step_definitions.append(get_matcher(func, step_text))

    def find_step_definition(self, step):
        candidates = self.steps[step.step_type]
        more_steps = self.steps["step"]
        if step.step_type != "step" and more_steps:
            # -- ENSURE: self.step_type lists are not modified/extended.
            candidates = list(candidates)
            candidates += more_steps

        for step_definition in candidates:
            if step_definition.match(step.name):
                return step_definition
        return None

    def find_match(self, step):
        candidates = self.steps[step.step_type]
        more_steps = self.steps["step"]
        if step.step_type != "step" and more_steps:
            # -- ENSURE: self.step_type lists are not modified/extended.
            candidates = list(candidates)
            candidates += more_steps

        for step_definition in candidates:
            result = step_definition.match(step.name)
            if result:
                return result

        return None

    def make_decorator(self, step_type):
        def decorator(step_text):
            def wrapper(func):
                self.add_step_definition(step_type, step_text, func)
                return func
            return wrapper
        return decorator


registry = StepRegistry()

# -- Create the decorators
# pylint: disable=redefined-outer-name
def setup_step_decorators(run_context=None, registry=registry):
    if run_context is None:
        run_context = globals()
    for step_type in ("given", "when", "then", "step"):
        step_decorator = registry.make_decorator(step_type)
        run_context[step_type.title()] = run_context[step_type] = step_decorator

# -----------------------------------------------------------------------------
# MODULE INIT:
# -----------------------------------------------------------------------------
setup_step_decorators()



# -*- coding: utf-8 -*-

from __future__ import absolute_import, print_function
import argparse
import inspect
import json
import logging
import os
import re
import sys
import shlex
import six
from six.moves import configparser

from behave.model import ScenarioOutline
from behave.model_core import FileLocation
from behave.reporter.junit import JUnitReporter
from behave.reporter.summary import SummaryReporter
from behave.tag_expression import make_tag_expression
from behave.formatter.base import StreamOpener
from behave.formatter import _registry as _format_registry
from behave.userdata import UserData, parse_user_define
from behave._types import Unknown
from behave.textutil import select_best_encoding, to_texts

# -- PYTHON 2/3 COMPATIBILITY:
# SINCE Python 3.2: ConfigParser = SafeConfigParser
ConfigParser = configparser.ConfigParser
if six.PY2:
    ConfigParser = configparser.SafeConfigParser

try:
    if sys.version_info >= (3, 11):
        import tomllib
    elif sys.version_info < (3, 0):
        import toml as tomllib
    else:
        import tomli as tomllib
    _TOML_AVAILABLE = True
except ImportError:
    _TOML_AVAILABLE = False


# -----------------------------------------------------------------------------
# CONSTANTS:
# -----------------------------------------------------------------------------
DEFAULT_RUNNER_CLASS_NAME = "behave.runner:Runner"


# -----------------------------------------------------------------------------
# CONFIGURATION DATA TYPES and TYPE CONVERTERS:
# -----------------------------------------------------------------------------
class LogLevel(object):
    names = [
        "NOTSET", "CRITICAL", "FATAL", "ERROR",
        "WARNING", "WARN", "INFO", "DEBUG",
    ]

    @staticmethod
    def parse(levelname, unknown_level=None):
        """
        Convert levelname into a numeric log level.

        :param levelname: Logging levelname (as string)
        :param unknown_level: Used if levelname is unknown (optional).
        :return: Numeric log-level or unknown_level, if levelname is unknown.
        """
        return getattr(logging, levelname.upper(), unknown_level)

    @classmethod
    def parse_type(cls, levelname):
        level = cls.parse(levelname, Unknown)
        if level is Unknown:
            message = "%s is unknown, use: %s" % \
                      (levelname, ", ".join(cls.names[1:]))
            raise argparse.ArgumentTypeError(message)
        return level

    @staticmethod
    def to_string(level):
        return logging.getLevelName(level)


def positive_number(text):
    """Converts a string into a positive integer number."""
    value = int(text)
    if value < 0:
        raise ValueError("POSITIVE NUMBER, but was: %s" % text)
    return value


# -----------------------------------------------------------------------------
# CONFIGURATION SCHEMA:
# -----------------------------------------------------------------------------
options = [
    (("-c", "--no-color"),
     dict(action="store_false", dest="color",
          help="Disable the use of ANSI color escapes.")),

    (("--color",),
     dict(dest="color", choices=["never", "always", "auto"],
          default=os.getenv('BEHAVE_COLOR'), const="auto", nargs="?",
          help="""Use ANSI color escapes. Defaults to %(const)r.
                  This switch is used to override a
                  configuration file setting.""")),

    (("-d", "--dry-run"),
     dict(action="store_true",
          help="Invokes formatters without executing the steps.")),

    (("-D", "--define"),
     dict(dest="userdata_defines", type=parse_user_define, action="append",
          metavar="NAME=VALUE",
          help="""Define user-specific data for the config.userdata dictionary.
                  Example: -D foo=bar to store it in config.userdata["foo"].""")),

    (("-e", "--exclude"),
     dict(metavar="PATTERN", dest="exclude_re",
          help="""Don't run feature files matching regular expression
                  PATTERN.""")),

    (("-i", "--include"),
     dict(metavar="PATTERN", dest="include_re",
          help="Only run feature files matching regular expression PATTERN.")),

    (("--no-junit",),
     dict(action="store_false", dest="junit",
          help="Don't output JUnit-compatible reports.")),

    (("--junit",),
     dict(action="store_true",
          help="""Output JUnit-compatible reports.
                  When junit is enabled, all stdout and stderr
                  will be redirected and dumped to the junit report,
                  regardless of the "--capture" and "--no-capture" options.
                  """)),

    (("--junit-directory",),
     dict(metavar="PATH", dest="junit_directory",
          default="reports",
          help="""Directory in which to store JUnit reports.""")),

    (("-j", "--jobs", "--parallel"),
     dict(metavar="NUMBER", dest="jobs", default=1, type=positive_number,
          help="""Number of concurrent jobs to use (default: %(default)s).
                  Only supported by test runners that support parallel execution.""")),

    ((),  # -- CONFIGFILE only
     dict(dest="default_format", default="pretty",
          help="Specify default formatter (default: %(default)s).")),


    (("-f", "--format"),
     dict(action="append",
          help="""Specify a formatter. If none is specified the default
                  formatter is used. Pass "--format help" to get a
                  list of available formatters.""")),

    (("--steps-catalog",),
     dict(action="store_true", dest="steps_catalog",
          help="""Show a catalog of all available step definitions.
                  SAME AS: --format=steps.catalog --dry-run --no-summary -q""")),

    ((),  # -- CONFIGFILE only
     dict(dest="scenario_outline_annotation_schema",
          help="""Specify name annotation schema for scenario outline
                  (default="{name} -- @{row.id} {examples.name}").""")),

    (("--no-skipped",),
     dict(action="store_false", dest="show_skipped",
          help="Don't print skipped steps (due to tags).")),

    (("--show-skipped",),
     dict(action="store_true",
          help="""Print skipped steps.
                  This is the default behaviour. This switch is used to
                  override a configuration file setting.""")),

    (("--no-snippets",),
     dict(action="store_false", dest="show_snippets",
          help="Don't print snippets for unimplemented steps.")),
    (("--snippets",),
     dict(action="store_true", dest="show_snippets",
          help="""Print snippets for unimplemented steps.
                  This is the default behaviour. This switch is used to
                  override a configuration file setting.""")),

    (("--no-multiline",),
     dict(action="store_false", dest="show_multiline",
          help="""Don't print multiline strings and tables under steps.""")),

    (("--multiline", ),
     dict(action="store_true", dest="show_multiline",
          help="""Print multiline strings and tables under steps.
                  This is the default behaviour. This switch is used to
                  override a configuration file setting.""")),

    (("-n", "--name"),
     dict(action="append", metavar="NAME_PATTERN",
          help="""Select feature elements (scenarios, ...) to run
                  which match part of the given name (regex pattern).
                  If this option is given more than once,
                  it will match against all the given names.""")),

    (("--no-capture",),
     dict(action="store_false", dest="stdout_capture",
          help="""Don't capture stdout (any stdout output will be
                  printed immediately.)""")),

    (("--capture",),
     dict(action="store_true", dest="stdout_capture",
          help="""Capture stdout (any stdout output will be
                  printed if there is a failure.)
                  This is the default behaviour. This switch is used to
                  override a configuration file setting.""")),

    (("--no-capture-stderr",),
     dict(action="store_false", dest="stderr_capture",
          help="""Don't capture stderr (any stderr output will be
                  printed immediately.)""")),

    (("--capture-stderr",),
     dict(action="store_true", dest="stderr_capture",
          help="""Capture stderr (any stderr output will be
                  printed if there is a failure.)
                  This is the default behaviour. This switch is used to
                  override a configuration file setting.""")),

    (("--no-logcapture",),
     dict(action="store_false", dest="log_capture",
          help="""Don't capture logging. Logging configuration will
                  be left intact.""")),

    (("--logcapture",),
     dict(action="store_true", dest="log_capture",
          help="""Capture logging. All logging during a step will be captured
                  and displayed in the event of a failure.
                  This is the default behaviour. This switch is used to
                  override a configuration file setting.""")),

    (("--logging-level",),
     dict(type=LogLevel.parse_type,
          help="""Specify a level to capture logging at. The default
                  is INFO - capturing everything.""")),

    (("--logging-format",),
     dict(help="""Specify custom format to print statements. Uses the
                  same format as used by standard logging handlers. The
                  default is "%%(levelname)s:%%(name)s:%%(message)s".""")),

    (("--logging-datefmt",),
     dict(help="""Specify custom date/time format to print
                  statements.
                  Uses the same format as used by standard logging
                  handlers.""")),

    (("--logging-filter",),
     dict(help="""Specify which statements to filter in/out. By default,
                  everything is captured. If the output is too verbose, use
                  this option to filter out needless output.
                  Example: --logging-filter=foo will capture statements issued
                  ONLY to foo or foo.what.ever.sub but not foobar or other
                  logger. Specify multiple loggers with comma:
                  filter=foo,bar,baz.
                  If any logger name is prefixed with a minus, eg filter=-foo,
                  it will be excluded rather than included.""",
          config_help="""Specify which statements to filter in/out. By default,
                         everything is captured. If the output is too verbose,
                         use this option to filter out needless output.
                         Example: ``logging_filter = foo`` will capture
                         statements issued ONLY to "foo" or "foo.what.ever.sub"
                         but not "foobar" or other logger. Specify multiple
                         loggers with comma: ``logging_filter = foo,bar,baz``.
                         If any logger name is prefixed with a minus, eg
                         ``logging_filter = -foo``, it will be excluded rather
                         than included.""")),

    (("--logging-clear-handlers",),
     dict(action="store_true",
          help="Clear all other logging handlers.")),

    (("--no-summary",),
     dict(action="store_false", dest="summary",
          help="""Don't display the summary at the end of the run.""")),

    (("--summary",),
     dict(action="store_true", dest="summary",
          help="""Display the summary at the end of the run.""")),

    (("-o", "--outfile"),
     dict(action="append", dest="outfiles", metavar="FILE",
          help="Write to specified file instead of stdout.")),

    ((),  # -- CONFIGFILE only
     dict(action="append", dest="paths",
          help="Specify default feature paths, used when none are provided.")),

    (("-q", "--quiet"),
     dict(action="store_true",
          help="Alias for --no-snippets --no-source.")),

    (("-r", "--runner"),
     dict(dest="runner", action="store", metavar="RUNNER_CLASS",
          default=DEFAULT_RUNNER_CLASS_NAME,
          help='Use own runner class, like: "behave.runner:Runner"')),

    (("--no-source",),
     dict(action="store_false", dest="show_source",
          help="""Don't print the file and line of the step definition with the
                  steps.""")),

    (("--show-source",),
     dict(action="store_true", dest="show_source",
          help="""Print the file and line of the step
                  definition with the steps. This is the default
                  behaviour. This switch is used to override a
                  configuration file setting.""")),

    (("--stage",),
     dict(help="""Defines the current test stage.
                  The test stage name is used as name prefix for the environment
                  file and the steps directory (instead of default path names).
                  """)),

    (("--stop",),
     dict(action="store_true",
          help="Stop running tests at the first failure.")),

    # -- DISABLE-UNUSED-OPTION: Not used anywhere.
    # (("-S", "--strict"),
    # dict(action="store_true",
    #    help="Fail if there are any undefined or pending steps.")),

    ((),  # -- CONFIGFILE only
     dict(dest="default_tags", metavar="TAG_EXPRESSION", action="append",
          help="""Define default tags when non are provided.
                  See --tags for more information.""")),

    (("-t", "--tags"),
     dict(action="append", metavar="TAG_EXPRESSION",
          help="""Only execute features or scenarios with tags
                  matching TAG_EXPRESSION. Pass "--tags-help" for
                  more information.""",
          config_help="""Only execute certain features or scenarios based
                         on the tag expression given. See below for how to code
                         tag expressions in configuration files.""")),

    (("-T", "--no-timings"),
     dict(action="store_false", dest="show_timings",
          help="""Don't print the time taken for each step.""")),

    (("--show-timings",),
     dict(action="store_true", dest="show_timings",
          help="""Print the time taken, in seconds, of each step after the
                  step has completed. This is the default behaviour. This
                  switch is used to override a configuration file
                  setting.""")),

    (("-v", "--verbose"),
     dict(action="store_true",
          help="Show the files and features loaded.")),

    (("-w", "--wip"),
     dict(action="store_true",
          help="""Only run scenarios tagged with "wip". Additionally: use the
                  "plain" formatter, do not capture stdout or logging output
                  and stop at the first failure.""")),

    (("--lang",),
     dict(metavar="LANG",
          help="Use keywords for a language other than English.")),

    (("--lang-list",),
     dict(action="store_true",
          help="List the languages available for --lang.")),

    (("--lang-help",),
     dict(metavar="LANG",
          help="List the translations accepted for one language.")),

    (("--tags-help",),
     dict(action="store_true",
          help="Show help for tag expressions.")),

    (("--version",),
     dict(action="store_true", help="Show version.")),
]

# -- OPTIONS: With raw value access semantics in configuration file.
raw_value_options = frozenset([
    "logging_format",
    "logging_datefmt",
    # -- MAYBE: "scenario_outline_annotation_schema",
])


def values_to_str(d):
    return json.loads(
        json.dumps(d),
        parse_float=str,
        parse_int=str,
        parse_constant=str
    )


def decode_options(config):
    for fixed, keywords in options:
        if "dest" in keywords:
            dest = keywords["dest"]
        else:
            dest = None
            for opt in fixed:
                if opt.startswith("--"):
                    dest = opt[2:].replace("-", "_")
                else:
                    assert len(opt) == 2
                    dest = opt[1:]
        if (
            not dest
        ) or (
            dest in "tags_help lang_list lang_help version".split()
        ):
            continue
        try:
            if dest not in config["behave"]:
                continue
        except AttributeError as exc:
            # SafeConfigParser instance has no attribute '__getitem__' (py27)
            if "__getitem__" not in str(exc):
                raise
            if not config.has_option("behave", dest):
                continue
        except KeyError:
            continue
        action = keywords.get("action", "store")
        yield dest, action


def format_outfiles_coupling(result, config_dir):
    # -- STEP: format/outfiles coupling
    if "format" in result:
        # -- OPTIONS: format/outfiles are coupled in configuration file.
        formatters = result["format"]
        formatter_size = len(formatters)
        outfiles = result.get("outfiles", [])
        outfiles_size = len(outfiles)
        if outfiles_size < formatter_size:
            for formatter_name in formatters[outfiles_size:]:
                outfile = "%s.output" % formatter_name
                outfiles.append(outfile)
            result["outfiles"] = outfiles
        elif len(outfiles) > formatter_size:
            print("CONFIG-ERROR: Too many outfiles (%d) provided." %
                  outfiles_size)
            result["outfiles"] = outfiles[:formatter_size]

    for paths_name in ("paths", "outfiles"):
        if paths_name in result:
            # -- Evaluate relative paths relative to location.
            # NOTE: Absolute paths are preserved by os.path.join().
            paths = result[paths_name]
            result[paths_name] = \
                [os.path.normpath(os.path.join(config_dir, p)) for p in paths]


def read_configparser(path):
    # pylint: disable=too-many-locals, too-many-branches
    config = ConfigParser()
    config.optionxform = str    # -- SUPPORT: case-sensitive keys
    config.read(path)
    config_dir = os.path.dirname(path)
    result = {}

    for dest, action in decode_options(config):
        if action == "store":
            result[dest] = config.get(
                "behave", dest, raw=dest in raw_value_options
            )
        elif action in ("store_true", "store_false"):
            result[dest] = config.getboolean("behave", dest)
        elif action == "append":
            if dest == "userdata_defines":
                continue    # -- SKIP-CONFIGFILE: Command-line only option.
            result[dest] = \
                [s.strip() for s in config.get("behave", dest).splitlines()]
        else:
            raise ValueError('action "%s" not implemented' % action)

    format_outfiles_coupling(result, config_dir)

    # -- STEP: Special additional configuration sections.
    # SCHEMA: config_section: data_name
    special_config_section_map = {
        "behave.formatters": "more_formatters",
        "behave.runners":    "more_runners",
        "behave.userdata":   "userdata",
    }
    for section_name, data_name in special_config_section_map.items():
        result[data_name] = {}
        if config.has_section(section_name):
            result[data_name].update(config.items(section_name))

    return result


def read_toml(path):
    """Read configuration from pyproject.toml file.

    Configuration should be stored inside the 'tool.behave' table.

    See https://www.python.org/dev/peps/pep-0518/#tool-table
    """
    # pylint: disable=too-many-locals, too-many-branches
    with open(path, "rb") as tomlfile:
        config = json.loads(json.dumps(tomllib.load(tomlfile)))  # simple dict

    config = config['tool']
    config_dir = os.path.dirname(path)
    result = {}

    for dest, action in decode_options(config):
        raw = config["behave"][dest]
        if action == "store":
            result[dest] = str(raw)
        elif action in ("store_true", "store_false"):
            result[dest] = bool(raw)
        elif action == "append":
            if dest == "userdata_defines":
                continue    # -- SKIP-CONFIGFILE: Command-line only option.
            # toml has native arrays and quoted strings, so there's no
            # need to split by newlines or strip values
            result[dest] = raw
        else:
            raise ValueError('action "%s" not implemented' % action)
    format_outfiles_coupling(result, config_dir)

    # -- STEP: Special additional configuration sections.
    # SCHEMA: config_section: data_name
    special_config_section_map = {
        "formatters": "more_formatters",
        "userdata":   "userdata",
    }
    for section_name, data_name in special_config_section_map.items():
        result[data_name] = {}
        try:
            result[data_name] = values_to_str(config["behave"][section_name])
        except KeyError:
            result[data_name] = {}

    return result


def read_configuration(path, verbose=False):
    ext = path.split(".")[-1]
    parsers = {
        "ini": read_configparser,
        "cfg": read_configparser,
        "behaverc": read_configparser,
    }

    if _TOML_AVAILABLE:
        parsers["toml"] = read_toml
    parse_func = parsers.get(ext, None)
    if not parse_func:
        if verbose:
            print('Unable to find a parser for "%s"' % path)
        return {}
    parsed = parse_func(path)

    return parsed


def config_filenames():
    paths = ["./", os.path.expanduser("~")]
    if sys.platform in ("cygwin", "win32") and "APPDATA" in os.environ:
        paths.append(os.path.join(os.environ["APPDATA"]))

    for path in reversed(paths):
        for filename in reversed((
            "behave.ini", ".behaverc", "setup.cfg", "tox.ini", "pyproject.toml"
        )):
            filename = os.path.join(path, filename)
            if os.path.isfile(filename):
                yield filename


def load_configuration(defaults, verbose=False):
    for filename in config_filenames():
        if verbose:
            print('Loading config defaults from "%s"' % filename)
        defaults.update(read_configuration(filename, verbose))

    if verbose:
        print("Using CONFIGURATION DEFAULTS:")
        for k, v in sorted(six.iteritems(defaults)):
            print("%18s: %s" % (k, v))


def setup_parser():
    # construct the parser
    # usage = "%(prog)s [options] [ [FILE|DIR|URL][:LINE[:LINE]*] ]+"
    usage = "%(prog)s [options] [ [DIR|FILE|FILE:LINE] ]+"
    description = """\
    Run a number of feature tests with behave."""
    more = """
    EXAMPLES:
    behave features/
    behave features/one.feature features/two.feature
    behave features/one.feature:10
    behave @features.txt
    """
    parser = argparse.ArgumentParser(usage=usage, description=description)
    for fixed, keywords in options:
        if not fixed:
            continue    # -- CONFIGFILE only.
        if "config_help" in keywords:
            keywords = dict(keywords)
            del keywords["config_help"]
        parser.add_argument(*fixed, **keywords)
    parser.add_argument("paths", nargs="*",
                        help="Feature directory, file or file location (FILE:LINE).")
    return parser


class Configuration(object):
    """Configuration object for behave and behave runners."""
    # pylint: disable=too-many-instance-attributes
    defaults = dict(
        color='never' if sys.platform == "win32" else os.getenv('BEHAVE_COLOR', 'auto'),
        jobs=1,
        show_snippets=True,
        show_skipped=True,
        dry_run=False,
        show_source=True,
        show_timings=True,
        stdout_capture=True,
        stderr_capture=True,
        log_capture=True,
        logging_format="%(levelname)s:%(name)s:%(message)s",
        logging_level=logging.INFO,
        runner=DEFAULT_RUNNER_CLASS_NAME,
        steps_catalog=False,
        summary=True,
        junit=False,
        stage=None,
        userdata={},
        # -- SPECIAL:
        default_format="pretty",    # -- Used when no formatters are configured.
        default_tags="",            # -- Used when no tags are defined.
        scenario_outline_annotation_schema=u"{name} -- @{row.id} {examples.name}"
    )
    cmdline_only_options = set("userdata_defines")

    def __init__(self, command_args=None, load_config=True, verbose=None,
                 **kwargs):
        """
        Constructs a behave configuration object.
          * loads the configuration defaults (if needed).
          * process the command-line args
          * store the configuration results

        :param command_args: Provide command args (as sys.argv).
            If command_args is None, sys.argv[1:] is used.
        :type command_args: list<str>, str
        :param load_config: Indicate if configfile should be loaded (=true)
        :param verbose: Indicate if diagnostic output is enabled
        :param kwargs:  Used to hand-over/overwrite default values.
        """
        # pylint: disable=too-many-branches, too-many-statements
        if command_args is None:
            command_args = sys.argv[1:]
        elif isinstance(command_args, six.string_types):
            encoding = select_best_encoding() or "utf-8"
            if six.PY2 and isinstance(command_args, six.text_type):
                command_args = command_args.encode(encoding)
            elif six.PY3 and isinstance(command_args, six.binary_type):
                command_args = command_args.decode(encoding)
            command_args = shlex.split(command_args)
        elif isinstance(command_args, (list, tuple)):
            command_args = to_texts(command_args)

        if verbose is None:
            # -- AUTO-DISCOVER: Verbose mode from command-line args.
            verbose = ("-v" in command_args) or ("--verbose" in command_args)

        # Allow commands like `--color features/whizbang.feature` to work
        # Without this, argparse will treat the positional arg as the value to
        # --color and we'd get:
        #   argument --color: invalid choice: 'features/whizbang.feature'
        #   (choose from 'never', 'always', 'auto')
        if '--color' in command_args:
            color_arg_pos = command_args.index('--color')
            if os.path.exists(command_args[color_arg_pos + 1]):
                command_args.insert(color_arg_pos + 1, '--')

        self.version = None
        self.tags_help = None
        self.lang_list = None
        self.lang_help = None
        self.default_tags = None
        self.junit = None
        self.logging_format = None
        self.logging_datefmt = None
        self.name = None
        self.scope = None
        self.steps_catalog = None
        self.userdata = None
        self.wip = None
        self.verbose = verbose

        defaults = self.defaults.copy()
        for name, value in six.iteritems(kwargs):
            defaults[name] = value
        self.defaults = defaults
        self.formatters = []
        self.reporters = []
        self.name_re = None
        self.outputs = []
        self.include_re = None
        self.exclude_re = None
        self.scenario_outline_annotation_schema = None  # pylint: disable=invalid-name
        self.steps_dir = "steps"
        self.environment_file = "environment.py"
        self.userdata_defines = None
        self.more_formatters = None
        self.more_runners = None
        self.runner_aliases = dict(default=DEFAULT_RUNNER_CLASS_NAME)
        if load_config:
            load_configuration(self.defaults, verbose=verbose)
        parser = setup_parser()
        parser.set_defaults(**self.defaults)
        args = parser.parse_args(command_args)
        for key, value in six.iteritems(args.__dict__):
            if key.startswith("_") and key not in self.cmdline_only_options:
                continue
            setattr(self, key, value)

        # -- ATTRIBUTE-NAME-CLEANUP:
        self.tag_expression = None
        self._tags = self.tags
        self.tags = None
        if isinstance(self.default_tags, six.string_types):
            self.default_tags = self.default_tags.split()

        self.paths = [os.path.normpath(path) for path in self.paths]
        self.setup_outputs(args.outfiles)

        if self.steps_catalog:
            # -- SHOW STEP-CATALOG: As step summary.
            self.default_format = "steps.catalog"
            if self.format:
                self.format.append("steps.catalog")
            else:
                self.format = ["steps.catalog"]
            self.dry_run = True
            self.summary = False
            self.show_skipped = False
            self.quiet = True

        if self.wip:
            # Only run scenarios tagged with "wip".
            # Additionally:
            #  * use the "plain" formatter (per default)
            #  * do not capture stdout or logging output and
            #  * stop at the first failure.
            self.default_format = "plain"
            self._tags = ["wip"] + self.default_tags
            self.color = False
            self.stop = True
            self.log_capture = False
            self.stdout_capture = False

        self.tag_expression = make_tag_expression(self._tags or self.default_tags)
        # -- BACKWARD-COMPATIBLE (BAD-NAMING STYLE; deprecating):
        self.tags = self.tag_expression

        if self.quiet:
            self.show_source = False
            self.show_snippets = False

        if self.exclude_re:
            self.exclude_re = re.compile(self.exclude_re)

        if self.include_re:
            self.include_re = re.compile(self.include_re)
        if self.name:
            # -- SELECT: Scenario-by-name, build regular expression.
            self.name_re = self.build_name_re(self.name)

        if self.stage is None:  # pylint: disable=access-member-before-definition
            # -- USE ENVIRONMENT-VARIABLE, if stage is undefined.
            self.stage = os.environ.get("BEHAVE_STAGE", None)
        self.setup_stage(self.stage)
        self.setup_model()
        self.setup_userdata()
        self.setup_runner_aliases()

        # -- FINALLY: Setup Reporters and Formatters
        # NOTE: Reporters and Formatters can now use userdata information.
        if self.junit:
            # Buffer the output (it will be put into Junit report)
            self.stdout_capture = True
            self.stderr_capture = True
            self.log_capture = True
            self.reporters.append(JUnitReporter(self))
        if self.summary:
            self.reporters.append(SummaryReporter(self))

        self.setup_formats()
        bad_formats_and_errors = self.select_bad_formats_with_errors()
        if bad_formats_and_errors:
            bad_format_parts = []
            for name, error in bad_formats_and_errors:
                message = "%s (problem: %s)" % (name, error)
                bad_format_parts.append(message)
            parser.error("BAD_FORMAT=%s" % ", ".join(bad_format_parts))

    def setup_outputs(self, args_outfiles=None):
        if self.outputs:
            assert not args_outfiles, "ONLY-ONCE"
            return

        # -- NORMAL CASE: Setup only initially (once).
        if not args_outfiles:
            self.outputs.append(StreamOpener(stream=sys.stdout))
        else:
            for outfile in args_outfiles:
                if outfile and outfile != "-":
                    self.outputs.append(StreamOpener(outfile))
                else:
                    self.outputs.append(StreamOpener(stream=sys.stdout))

    def setup_formats(self):
        """Register more, user-defined formatters by name."""
        if self.more_formatters:
            for name, scoped_class_name in self.more_formatters.items():
                _format_registry.register_as(name, scoped_class_name)

    def setup_runner_aliases(self):
        if self.more_runners:
            for name, scoped_class_name in self.more_runners.items():
                self.runner_aliases[name] = scoped_class_name

    def select_bad_formats_with_errors(self):
        bad_formats = []
        if self.format:
            for format_name in self.format:
                formatter_valid = _format_registry.is_formatter_valid(format_name)
                if format_name == "help" or formatter_valid:
                    continue

                try:
                    _ = _format_registry.select_formatter_class(format_name)
                    bad_formats.append((format_name, "InvalidClassError"))
                except Exception as e:
                    formatter_error = e.__class__.__name__
                    if formatter_error == "KeyError":
                        formatter_error = "LookupError"
                    if self.verbose:
                        formatter_error += ": %s" % str(e)
                    bad_formats.append((format_name, formatter_error))
        return bad_formats

    @staticmethod
    def build_name_re(names):
        """
        Build regular expression for scenario selection by name
        by using a list of name parts or name regular expressions.

        :param names: List of name parts or regular expressions (as text).
        :return: Compiled regular expression to use.
        """
        # -- NOTE: re.LOCALE is removed in Python 3.6 (deprecated in Python 3.5)
        # flags = (re.UNICODE | re.LOCALE)
        # -- ENSURE: Names are all unicode/text values (for issue #606).
        names = to_texts(names)
        pattern = u"|".join(names)
        return re.compile(pattern, flags=re.UNICODE)

    def exclude(self, filename):
        if isinstance(filename, FileLocation):
            filename = six.text_type(filename)

        if self.include_re and self.include_re.search(filename) is None:
            return True
        if self.exclude_re and self.exclude_re.search(filename) is not None:
            return True
        return False

    def setup_logging(self, level=None, configfile=None, **kwargs):
        """
        Support simple setup of logging subsystem.
        Ensures that the logging level is set.
        But note that the logging setup can only occur once.

        SETUP MODES:
          * :func:`logging.config.fileConfig()`, if ``configfile`` is provided.
          * :func:`logging.basicConfig()`, otherwise.

        .. code-block: python
            # -- FILE: features/environment.py
            def before_all(context):
                context.config.setup_logging()

        :param level:       Logging level of root logger.
                            If None, use :attr:`logging_level` value.
        :param configfile:  Configuration filename for fileConfig() setup.
        :param kwargs:      Passed to :func:`logging.basicConfig()`
        """
        if level is None:
            level = self.logging_level      # pylint: disable=no-member

        if configfile:
            from logging.config import fileConfig
            fileConfig(configfile)
        else:
            # pylint: disable=no-member
            format_ = kwargs.pop("format", self.logging_format)
            datefmt = kwargs.pop("datefmt", self.logging_datefmt)
            logging.basicConfig(format=format_, datefmt=datefmt, **kwargs)
        # -- ENSURE: Default log level is set
        #    (even if logging subsystem is already configured).
        logging.getLogger().setLevel(level)

    def setup_model(self):
        if self.scenario_outline_annotation_schema:
            name_schema = six.text_type(self.scenario_outline_annotation_schema)
            ScenarioOutline.annotation_schema = name_schema.strip()

    def setup_stage(self, stage=None):
        """Setup the test stage that selects a different set of
        steps and environment implementations.

        :param stage:   Name of current test stage (as string or None).

        EXAMPLE::

            # -- SETUP DEFAULT TEST STAGE (unnamed):
            config = Configuration()
            config.setup_stage()
            assert config.steps_dir == "steps"
            assert config.environment_file == "environment.py"

            # -- SETUP PRODUCT TEST STAGE:
            config.setup_stage("product")
            assert config.steps_dir == "product_steps"
            assert config.environment_file == "product_environment.py"
        """
        steps_dir = "steps"
        environment_file = "environment.py"
        if stage:
            # -- USE A TEST STAGE: Select different set of implementations.
            prefix = stage + "_"
            steps_dir = prefix + steps_dir
            environment_file = prefix + environment_file
        self.steps_dir = steps_dir
        self.environment_file = environment_file

    def setup_userdata(self):
        if not isinstance(self.userdata, UserData):
            self.userdata = UserData(self.userdata)
        if self.userdata_defines:
            # -- ENSURE: Cmd-line overrides configuration file parameters.
            self.userdata.update(self.userdata_defines)

    def update_userdata(self, data):
        """Update userdata with data and reapply userdata defines (cmdline).
        :param data:  Provides (partial) userdata (as dict)
        """
        self.userdata.update(data)
        if self.userdata_defines:
            # -- REAPPLY: Cmd-line defines (override configuration file data).
            self.userdata.update(self.userdata_defines)



# -*- coding: UTF-8 -*-
"""
Contains classes and functionality to provide the active-tag mechanism.
Active-tags provide a skip-if logic based on tags in feature files.
"""

from __future__ import absolute_import, print_function
import logging
import operator
import re
import six
from ._types import Unknown
from .compat.collections import UserDict


# -----------------------------------------------------------------------------
# VALUE OBJECT CLASSES FOR: Active-Tag Value Providers
# -----------------------------------------------------------------------------
class ValueObject(object):
    """Value object for active-tags that holds the current value for
    one activate-tag category and its comparison function.

    The :param:`compare_func(current_value, tag_value)` is a predicate function
    with two arguments that performs the comparison between the
    "current_value" and the "tag_value".

    EXAMPLE::

        # -- SIMPLIFIED EXAMPLE:
        from behave.tag_matcher import ValueObject
        import operator     # Module contains comparison functions.
        class NumberObject(ValueObject): ...  # Details left out here.

        xxx_current_value = 42
        active_tag_value_provider = {
            "xxx.value": ValueObject(xxx_current_value)  # USES: operator.eq (equals)
            "xxx.min_value": NumberValueObject(xxx_current_value, operator.ge),
            "xxx.max_value": NumberValueObject(xxx_current_value, operator.le),
        }

        # -- LATER WITHIN: ActivTag Logic
        # EXAMPLE TAG: @use.with_xxx.min_value=10  (schema: "@use.with_{category}={value}")
        tag_category = "xxx.min_value"
        current_value = active_tag_value_provider.get(tag_category)
        if not isinstance(current_value, ValueObject):
            current_value = ValueObject(current_value)
        ...
        tag_matches = current_value.matches(tag_value)
    """
    def __init__(self, value, compare=operator.eq):
        assert callable(compare)
        self._value = value
        self.compare = compare

    @property
    def value(self):
        if callable(self._value):
            # -- SUPPORT: Lazy computation of current-value.
            return self._value()
        # -- OTHERWISE:
        return self._value

    def matches(self, tag_value):
        """Comparison between current value and :param:`tag_value`.

        :param tag_value: Tag value from active tag (as string).
        :return: True, if comparison matches. False, otherwise.
        """
        return bool(self.compare(self.value, tag_value))

    @staticmethod
    def on_type_conversion_error(tag_value, e):
        logger = logging.getLogger("behave.active_tags")
        logger.error("TYPE CONVERSION ERROR: active_tag.value='%s' (error: %s)" % \
                     (tag_value, str(e)))
        # MAYBE: logger.exception(e)
        return False    # HINT: mis-matched

    def __str__(self):
        """Conversion to string."""
        return str(self.value)

    def __repr__(self):
        return "<%s: value=%s, compare=%s>" % \
               (self.__class__.__name__, self.value, self.compare)


class NumberValueObject(ValueObject):
    def matches(self, tag_value):
        try:
            tag_number = int(tag_value)
            return super(NumberValueObject, self).matches(tag_number)
        except ValueError as e:
            # -- INTEGER TYPE-CONVERSION ERROR:
            return self.on_type_conversion_error(tag_value, e)

    def __int__(self):
        """Convert into integer-number value."""
        return int(self.value)


class BoolValueObject(ValueObject):
    TRUE_STRINGS = set(["true", "yes", "on"])
    FALSE_STRINGS = set(["false", "no", "off"])

    def matches(self, tag_value):
        try:
            boolean_tag_value = self.to_bool(tag_value)
            return super(BoolValueObject, self).matches(boolean_tag_value)
        except ValueError as e:
            return self.on_type_conversion_error(tag_value, e)

    def __bool__(self):
        """Conversion to boolean value."""
        return bool(self.value)

    @classmethod
    def to_bool(cls, value):
        if isinstance(value, six.string_types):
            text = value.lower()
            if text in cls.TRUE_STRINGS:
                return True
            elif text in cls.FALSE_STRINGS:
                return False
            else:
                raise ValueError("NON-BOOL: %s" % value)
        # -- OTHERWISE:
        return bool(value)


# -----------------------------------------------------------------------------
# CLASSES FOR: Active-Tags and ActiveTagMatchers
# -----------------------------------------------------------------------------
class TagMatcher(object):
    """Abstract base class that defines the TagMatcher protocol."""

    def should_run_with(self, tags):
        """Determines if a feature/scenario with these tags should run or not.

        :param tags:    List of scenario/feature tags to check.
        :return: True,  if scenario/feature should run.
        :return: False, if scenario/feature should be excluded from the run-set.
        """
        return not self.should_exclude_with(tags)

    def should_exclude_with(self, tags):
        """Determines if a feature/scenario with these tags should be excluded
        from the run-set.

        :param tags:    List of scenario/feature tags to check.
        :return: True, if scenario/feature should be excluded from the run-set.
        :return: False, if scenario/feature should run.
        """
        raise NotImplementedError


class ActiveTagMatcher(TagMatcher):
    """Provides an active tag matcher for many categories.

    TAG SCHEMA 1 (preferred):
      * use.with_{category}={value}
      * not.with_{category}={value}

    TAG SCHEMA 2:
      * active.with_{category}={value}
      * not_active.with_{category}={value}

    TAG LOGIC
    ----------

    Determine active-tag groups by grouping active-tags
    with same category together::

        active_group.enabled := enabled(group.tag1) or enabled(group.tag2) or ...
        active_tags.enabled  := enabled(group1) and enabled(group2) and ...

    All active-tag groups must be turned "on" (enabled).
    Otherwise, the model element should be excluded.

    CONCEPT: ValueProvider
    ------------------------------

    A ValueProvider provides the value of a category, used in active tags.
    A ValueProvider must provide a mapping-like protocol:

    .. code-block:: python

        class MyValueProvider(object):
            def get(self, category_name, default=None):
                ...
                return category_value   # OR: default, if category is unknown.

    EXAMPLE:
    --------

    Run some scenarios only when runtime conditions are met:

      * Run scenario Alice only on Windows OS
      * Run scenario Bob with all browsers except Chrome

    .. code-block:: gherkin

        # -- FILE: features/alice.feature
        Feature:

          @use.with_os=win32
          Scenario: Alice (Run only on Windows)
            Given I do something
            ...

          @not.with_browser=chrome
          Scenario: Bob (Excluded with Web-Browser Chrome)
            Given I do something else
            ...


    .. code-block:: python

        # -- FILE: features/environment.py
        from behave.tag_matcher import ActiveTagMatcher
        import sys

        # -- MATCHES ANY ACTIVE TAGS: @{prefix}.with_{category}={value}
        # NOTE: active_tag_value_provider provides current category values.
        active_tag_value_provider = {
            "browser": os.environ.get("BEHAVE_BROWSER", "chrome"),
            "os":      sys.platform,
        }
        active_tag_matcher = ActiveTagMatcher(active_tag_value_provider)

        def before_feature(context, feature):
            if active_tag_matcher.should_exclude_with(feature.tags):
                feature.skip()   #< LATE-EXCLUDE from run-set.

        def before_scenario(context, scenario):
            if active_tag_matcher.should_exclude_with(scenario.effective_tags):
                exclude_reason = active_tag_matcher.exclude_reason
                scenario.skip(exclude_reason)   #< LATE-EXCLUDE from run-set.
    """
    value_separator = "="
    tag_prefixes = ["use", "not", "active", "not_active", "only"]
    tag_schema = r"^(?P<prefix>%s)\.with_(?P<category>\w+(\.\w+)*)%s(?P<value>.*)$"
    ignore_unknown_categories = True
    use_exclude_reason = False

    def __init__(self, value_provider, tag_prefixes=None,
                 value_separator=None, ignore_unknown_categories=None):
        if value_provider is None:
            value_provider = {}
        if tag_prefixes is None:
            tag_prefixes = self.tag_prefixes
        if ignore_unknown_categories is None:
            ignore_unknown_categories = self.ignore_unknown_categories

        super(ActiveTagMatcher, self).__init__()
        self.value_provider = value_provider
        self.tag_pattern = self.make_tag_pattern(tag_prefixes, value_separator)
        self.tag_prefixes = tag_prefixes
        self.ignore_unknown_categories = ignore_unknown_categories
        self.exclude_reason = None

    @classmethod
    def make_tag_pattern(cls, tag_prefixes, value_separator=None):
        if value_separator is None:
            value_separator = cls.value_separator
        any_tag_prefix = r"|".join(tag_prefixes)
        expression = cls.tag_schema % (any_tag_prefix, value_separator)
        return re.compile(expression)

    @classmethod
    def make_category_tag(cls, category, value=None,
                          tag_prefix=None, value_sep=None):
        """Build category tag (mostly for testing purposes).
        :return: Category tag as string (without leading AT char).
        """
        if tag_prefix is None:
            tag_prefix = cls.tag_prefixes[0]    # -- USE: First as default.
        if value_sep is None:
            value_sep = cls.value_separator
        value = value or ""
        return "%s.with_%s%s%s" % (tag_prefix, category, value_sep, value)

    def is_tag_negated(self, tag):      # pylint: disable=no-self-use
        return tag.startswith("not")

    def is_tag_group_enabled(self, group_category, group_tag_pairs):
        """Provides boolean logic to determine if all active-tags
        which use the same category result in an enabled value.

        .. code-block:: gherkin

            @use.with_xxx=alice
            @use.with_xxx=bob
            @not.with_xxx=charly
            @not.with_xxx=doro
            Scenario:
                Given a step passes
                ...

        Use LOGICAL expression for active-tags with same category::

            category_tag_group.enabled := positive-tag-expression and not negative-tag-expression
              positive-tag-expression  := enabled(tag1) or enabled(tag2) or ...
              negative-tag-expression  := enabled(tag3) or enabled(tag4) or ...
               tag1, tag2 are positive-tags, like @use.with_category=value
               tag3, tag4 are negative-tags, like @not.with_category=value

             xxx   | Only use parts: (xxx == "alice") or (xxx == "bob")
            -------+-------------------
            alice  | true
            bob    | true
            other  | false

             xxx   | Only not parts:
                   | (not xxx == "charly") and (not xxx == "doro")
                   | = not((xxx == "charly") or (xxx == "doro"))
            -------+-------------------
            charly | false
            doro   | false
            other  | true

             xxx   | Use and not parts:
                   | ((xxx == "alice") or (xxx == "bob")) and not((xxx == "charly") or (xxx == "doro"))
            -------+-------------------
            alice  | true
            bob    | true
            charly | false
            doro   | false
            other  | false

        :param group_category:      Category for this tag-group (as string).
        :param category_tag_group:  List of active-tag match-pairs.
        :return: True, if tag-group is enabled.
        """
        if not group_tag_pairs:
            # -- CASE: Empty group is always enabled (CORNER-CASE).
            return True

        current_value = self.value_provider.get(group_category, Unknown)
        if current_value is Unknown and self.ignore_unknown_categories:
            # -- CASE: Unknown category, ignore it.
            return True
        elif not isinstance(current_value, ValueObject):
            current_value = ValueObject(current_value)

        positive_tags_matched = []
        negative_tags_matched = []
        for category_tag, tag_match in group_tag_pairs:
            tag_prefix = tag_match.group("prefix")
            category = tag_match.group("category")
            tag_value = tag_match.group("value")
            assert category == group_category

            if self.is_tag_negated(tag_prefix):
                # -- CASE: @not.with_CATEGORY=VALUE
                # NORMALLY: tag_matched = (current_value == tag_value)
                tag_matched = current_value.matches(tag_value)
                negative_tags_matched.append(tag_matched)
            else:
                # -- CASE: @use.with_CATEGORY=VALUE
                # NORMALLY: tag_matched = (current_value == tag_value)
                tag_matched = current_value.matches(tag_value)
                positive_tags_matched.append(tag_matched)
        tag_expression1 = any(positive_tags_matched)    #< LOGICAL-OR expression
        tag_expression2 = any(negative_tags_matched)    #< LOGICAL-OR expression
        if not positive_tags_matched:
            tag_expression1 = True
        tag_group_enabled = bool(tag_expression1 and not tag_expression2)
        return tag_group_enabled

    def should_exclude_with(self, tags):
        group_categories = self.group_active_tags_by_category(tags)
        for group_category, category_tag_pairs in group_categories:
            if not self.is_tag_group_enabled(group_category, category_tag_pairs):
                # -- LOGICAL-AND SHORTCUT: Any false => Makes everything false
                if self.use_exclude_reason:
                    current_value = self.value_provider.get(group_category, None)
                    reason = "%s (but: %s)" % (group_category, current_value)
                    self.exclude_reason = reason
                return True     # SHOULD-EXCLUDE: not enabled = not False
        # -- LOGICAL-AND: All parts are True
        return False    # SHOULD-EXCLUDE: not enabled = not True

    def select_active_tags(self, tags):
        """Select all active tags that match the tag schema pattern.

        :param tags: List of tags (as string).
        :return: List of (tag, match_object) pairs (as generator).
        """
        for tag in tags:
            match_object = self.tag_pattern.match(tag)
            if match_object:
                yield (tag, match_object)

    def group_active_tags_by_category(self, tags):
        """Select all active tags that match the tag schema pattern
        and returns groups of active-tags, each group with tags
        of the same category.

        :param tags: List of tags (as string).
        :return: List of tag-groups (as generator), each tag-group is a
                list of (tag1, match1) pairs for the same category.
        """
        category_tag_groups = {}
        for tag in tags:
            match_object = self.tag_pattern.match(tag)
            if match_object:
                category = match_object.group("category")
                category_tag_pairs = category_tag_groups.get(category, None)
                if category_tag_pairs is None:
                    category_tag_pairs = category_tag_groups[category] = []
                category_tag_pairs.append((tag, match_object))

        for category, category_tag_pairs in six.iteritems(category_tag_groups):
            yield (category, category_tag_pairs)


class PredicateTagMatcher(TagMatcher):
    def __init__(self, exclude_function):
        assert callable(exclude_function)
        super(PredicateTagMatcher, self).__init__()
        self.predicate = exclude_function

    def should_exclude_with(self, tags):
        return self.predicate(tags)


class CompositeTagMatcher(TagMatcher):
    """Provides a composite tag matcher."""

    def __init__(self, tag_matchers=None):
        super(CompositeTagMatcher, self).__init__()
        self.tag_matchers = tag_matchers or []

    def should_exclude_with(self, tags):
        for tag_matcher in self.tag_matchers:
            if tag_matcher.should_exclude_with(tags):
                return True
        # -- OTHERWISE:
        return False


# -----------------------------------------------------------------------------
# ACTIVE TAG VALUE PROVIDER CLASSES:
# -----------------------------------------------------------------------------
class IActiveTagValueProvider(object):
    """Protocol/Interface for active-tag value providers."""

    def get(self, category, default=None):
        return NotImplemented


class ActiveTagValueProvider(UserDict):
    def __init__(self, data=None):
        if data is None:
            data = {}
        UserDict.__init__(self, data)

    @staticmethod
    def use_value(value):
        if callable(value):
            # -- RE-EVALUATE VALUE: Each time
            value_func = value
            value = value_func()
        return value

    def __getitem__(self, name):
        value = self.data[name]
        return self.use_value(value)

    def get(self, category, default=None):
        value = self.data.get(category, default)
        return self.use_value(value)

    def values(self):
        for value in self.data.values(self):
            yield self.use_value(value)

    def items(self):
        for category, value in self.data.items():
            yield (category, self.use_value(value))

    def categories(self):
        return self.keys()


class CompositeActiveTagValueProvider(ActiveTagValueProvider):
    """Provides a composite helper class to resolve active-tag values
    from a list of value-providers.
    """

    def __init__(self, value_providers=None):
        if value_providers is None:
            value_providers = []
        super(CompositeActiveTagValueProvider, self).__init__()
        self.value_providers = list(value_providers)

    def get(self, category, default=None):
        # -- FIRST: Check category cached-map (=self.data)
        value = self.data.get(category, Unknown)
        if value is Unknown:
            # -- NOT DISCOVERED: Search over value_providers.
            for value_provider in self.value_providers:
                value = value_provider.get(category, Unknown)
                if value is Unknown:
                    continue

                # -- FOUND CATEGORY:
                self.data[category] = value
                break
            # -- FOUND-CATEGORY or NOT-FOUND:
            if value is Unknown:
                value = default

        return self.use_value(value)

    # -- MORE: Provide a dict-like interface.
    def keys(self):
        for value_provider in self.value_providers:
            try:
                for category in value_provider.keys():
                    yield category
            except AttributeError:
                # -- keys() method not supported.
                pass

    def values(self):
        for category in self.keys():
            value = self.get(category)
            yield value

    def items(self):
        for category in self.keys():
            value = self.get(category)
            yield category, value


# -----------------------------------------------------------------------------
# UTILITY FUNCTIONS:
# -----------------------------------------------------------------------------
def bool_to_string(value):
    """Converts a boolean active-tag value into its normalized
    string representation.

    :param value:  Boolean value to use (or value converted into bool).
    :returns: Boolean value converted into a normalized string.
    """
    return str(bool(value)).lower()


def setup_active_tag_values(active_tag_values, data):
    """Setup/update active_tag values with dict-like data.
    Only values for keys that are already present are updated.

    :param active_tag_values:   Data storage for active_tag value (dict-like).
    :param data:   Data that should be used for active_tag values (dict-like).
    """
    for category in list(active_tag_values.keys()):
        if category in data:
            active_tag_values[category] = data[category]


def print_active_tags(active_tag_value_provider, categories=None):
    """Print a summary of the current active-tag values."""
    if categories is None:
        try:
            categories = list(active_tag_value_provider.keys())
        except TypeError:   # TypeError: object is not iterable
            categories = []

    active_tag_data = active_tag_value_provider
    print("ACTIVE-TAGS:")
    for category in categories:
        active_tag_value = active_tag_data.get(category)
        print("use.with_{category}={value}".format(
            category=category, value=active_tag_value))

    # -- FINALLY: TRAILING NEW-LINE
    print()



# -*- coding: UTF-8 -*-
"""
Create a runner as behave plugin by using its name:

* scoped-class-name, like: "behave.runner:Runner" (dotted.module:ClassName)
* runner-alias (alias mapping provided in config-file "behave.runners" section)

.. code-block:: ini

    # -- FILE: behave.ini
    # RUNNER-ALIAS EXAMPLE:
    # USE: behave --runner=default features/
    [behave.runners]
    default = behave.runner:Runner
"""

from __future__ import absolute_import, print_function
import inspect
from behave.api.runner import ITestRunner
from behave.exception import ConfigError, ClassNotFoundError, InvalidClassError, ModuleNotFoundError
from behave.importer import load_module, make_scoped_class_name, parse_scoped_name
from behave._types import Unknown


class RunnerPlugin(object):
    """Extension point to load a runner_class and create its runner:

    * create a runner by using its scoped-class-name
    * create a runner by using its runner-alias (provided in config-file)
    * create a runner by using a runner-class

    .. code-block:: python

        # -- EXAMPLE: Provide own test runner-class
        from behave.api.runner import ITestRunner

        class MyRunner(ITestRunner):
            def __init__(self, config, **kwargs):
                self.config = config

            def run(self):
                ... # IMPLEMENTATION DETAILS: Left out here.

    .. code-block:: python

        # -- CASE 1A: Make a runner by using its scoped-class-name
        plugin = RunnerPlugin("behave.runner:Runner")
        runner = plugin.make_runner(config)

        # -- CASE 1B: Make a runner by using its runner-alias
        # CONFIG-FILE SECTION: "behave.ini"
        #   [behave.runners]
        #   one = behave.runner:Runner
        plugin = RunnerPlugin("one")
        runner = plugin.make_runner(config)

        # -- CASE 2: Make a runner by using a runner-class
        from behave.runner import Runner as DefaultRunner
        plugin = RunnerPlugin(runner_class=DefaultRunner)
        runner = plugin.make_runner(config)
    """
    def __init__(self, runner_name=None, runner_class=None, runner_aliases=None):
        if not runner_class and runner_name and inspect.isclass(runner_name):
            # -- USABILITY: Use runner_name as runner_class
            runner_class = runner_name
            runner_name = make_scoped_class_name(runner_class)
        self.runner_name = runner_name
        self.runner_class = runner_class
        self.runner_aliases = runner_aliases or {}

    @staticmethod
    def is_class_valid(runner_class):
        run_method = getattr(runner_class, "run", None)
        return (inspect.isclass(runner_class) and
                issubclass(runner_class, ITestRunner) and
                callable(run_method))

    @staticmethod
    def validate_class(runner_class, runner_class_name=None):
        """Check if a runner class supports the Runner API constraints."""
        if not runner_class_name:
            runner_class_name = make_scoped_class_name(runner_class)

        if not inspect.isclass(runner_class):
            raise InvalidClassError("is not a class")
        elif not issubclass(runner_class, ITestRunner):
            message = "is not a subclass-of 'behave.api.runner:ITestRunner'"
            raise InvalidClassError(message)

        run_method = getattr(runner_class, "run", None)
        if not callable(run_method):
            raise InvalidClassError("run() is not callable")

        undefined_steps = getattr(runner_class, "undefined_steps", None)
        if undefined_steps is None:
            raise InvalidClassError("undefined_steps: missing attribute or property")
        # MAYBE:
        # elif not callable(undefined_steps):
        #     raise InvalidClassError("undefined_steps is not callable")

    @classmethod
    def load_class(cls, runner_class_name, verbose=True):
        """Loads a runner class by using its scoped-class-name, like:
        `my.module:Class1`.

        :param runner_class_name:  Scoped class-name (as string).
        :return: Loaded runner-class (on success).
        :raises ModleNotFoundError: If module does not exist or not importable.
        :raises ClassNotFoundError: If module exist, but class was not found.
        :raises InvalidClassError: If class is invalid (wrong subclass or not a class).
        :raises ImportError: If module was not found (or other Import-Errors above).
        """
        module_name, class_name = parse_scoped_name(runner_class_name)
        try:
            module = load_module(module_name)
            runner_class = getattr(module, class_name, Unknown)
            if runner_class is Unknown:
                raise ClassNotFoundError(runner_class_name)
            cls.validate_class(runner_class, runner_class_name)
            return runner_class
        except (ImportError, InvalidClassError, TypeError) as e:
            # -- CASE: ModuleNotFoundError, ClassNotFoundError, InvalidClassError, ...
            if verbose:
                print("BAD_RUNNER_CLASS: FAILED to load runner.class=%s (%s)" % \
                      (runner_class_name, e.__class__.__name__))
            raise

    @classmethod
    def make_problem_description(cls, scoped_class_name, use_details=False):
        """Check runner class for problems.

        :param scoped_class_name:  Runner class name (as string).
        :return: EMPTY_STRING, if no problem exists.
        :return: Problem exception class name (as string).
        """
        # -- STEP: Check runner for problems
        problem_description = ""
        try:
            cls.load_class(scoped_class_name, verbose=False)
        except (ImportError, TypeError) as e:
            problem_description = e.__class__.__name__
            if use_details:
                problem_description = "%s: %s" % (problem_description, str(e))
        return problem_description

    def make_runner(self, config, **runner_kwargs):
        """Build a runner either by:

        * providing a runner-class
        * using its name (alias-name or scoped-class-name).

        :param config:   Configuration object to use for runner.
        :param runner_kwargs:  Keyword args for runner creation.
        :return: Runner object to use.
        :raises ClassNotFoundError: If module exist, but class was not found.
        :raises InvalidClassError: If class is invalid (wrong subclass or not a class).
        :raises ImportError: If module was not found (or other Import-Errors above).
        :raises ConfigError: If runner-alias is not in config-file (section: behave.runners).
        """
        runner_class = self.runner_class
        if not runner_class:
            # -- CASE: Using runner-name (alias) or scoped_class_name.
            runner_name = self.runner_name or config.runner
            runner_aliases = {}
            runner_aliases.update(config.runner_aliases)
            runner_aliases.update(self.runner_aliases)
            scoped_class_name = runner_aliases.get(runner_name, runner_name)
            if scoped_class_name == runner_name and ":" not in scoped_class_name:
                # -- CASE: runner-alias is not in config-file section="behave.runner".
                raise ConfigError("runner=%s (RUNNER-ALIAS NOT FOUND)" % scoped_class_name)
            runner_class = self.load_class(scoped_class_name)
        else:
            self.validate_class(runner_class)

        runner = runner_class(config, **runner_kwargs)
        return runner



# -*- coding: UTF-8 -*-
"""
Provides a summary after each test run.
"""

from __future__ import absolute_import, division, print_function
import sys
from time import time as time_now
from behave.model import Rule, ScenarioOutline  # MAYBE: Scenario
from behave.model_core import Status
from behave.reporter.base import Reporter
from behave.formatter.base import StreamOpener


# ---------------------------------------------------------------------------
# CONSTANTS:
# ---------------------------------------------------------------------------
# -- DISABLED: OPTIONAL_STEPS = ('untested', 'undefined')
OPTIONAL_STEPS = (Status.untested,) # MAYBE: Status.undefined
STATUS_ORDER = (Status.passed, Status.failed, Status.skipped,
                Status.undefined, Status.untested)


# ---------------------------------------------------------------------------
# UTILITY FUNCTIONS:
# ---------------------------------------------------------------------------
def pluralize(word, count=1, suffix="s"):
    if count == 1:
        return word
    # -- OTHERWISE:
    return "{0}{1}".format(word, suffix)


def compute_summary_sum(summary):
    """Compute sum of all summary counts (except: all)

    :param summary: Summary counts (as dict).
    :return: Sum of all counts (as integer).
    """
    counts_sum = 0
    for name, count in summary.items():
        if name == "all":
            continue    # IGNORE IT.
        counts_sum += count
    return counts_sum


def format_summary0(statement_type, summary):
    parts = []
    for status in STATUS_ORDER:
        if status.name not in summary:
            continue
        counts = summary[status.name]
        if status in OPTIONAL_STEPS and counts == 0:
            # -- SHOW-ONLY: For relevant counts, suppress: untested items, etc.
            continue

        if not parts:
            # -- FIRST ITEM: Add statement_type to counter.
            label = statement_type
            if counts != 1:
                label += 's'
            part = u"%d %s %s" % (counts, label, status.name)
        else:
            part = u"%d %s" % (counts, status.name)
        parts.append(part)
    return ", ".join(parts) + "\n"


def format_summary(statement_type, summary):
    parts = []
    for status in STATUS_ORDER:
        if status.name not in summary:
            continue
        counts = summary[status.name]
        if status in OPTIONAL_STEPS and counts == 0:
            # -- SHOW-ONLY: For relevant counts, suppress: untested items, etc.
            continue

        name = status.name
        if status.name == "passed":
            statement = pluralize(statement_type, counts)
            name = u"%s passed" % statement
        part = u"%d %s" % (counts, name)
        parts.append(part)
    return ", ".join(parts) + "\n"


# -- PREPARED:
def format_summary2(statement_type, summary, end="\n"):
    """Format the summary line for one statement type.

    .. code-block::

        6 scenarios (passed: 5, failed: 1, skipped: 0, untested: 0)

    :param statement_type:
    :param summary:
    :return:
    """
    parts = []
    for status in STATUS_ORDER:
        if status.name not in summary:
            continue
        counts = summary[status.name]
        if status in OPTIONAL_STEPS and counts == 0:
            # -- SHOW-ONLY: For relevant counts, suppress: untested items, etc.
            continue
        parts.append((status.name, counts))

    counts_sum = summary["all"]
    statement = pluralize(statement_type, sum)
    parts_text = ", ".join(["{0}: {1}".format(name, value)
                            for name, value in parts])
    return "{count:4} {statement:<9} ({parts}){end}".format(
        count=counts_sum, statement=statement, parts=parts_text, end=end)


# ---------------------------------------------------------------------------
# REPORTERS:
# ---------------------------------------------------------------------------
class SummaryReporter(Reporter):
    show_failed_scenarios = True
    output_stream_name = "stdout"

    def __init__(self, config):
        super(SummaryReporter, self).__init__(config)
        stream = getattr(sys, self.output_stream_name, sys.stderr)
        self.stream = StreamOpener.ensure_stream_with_encoder(stream)
        summary_zero_data = {
            "all": 0,
            Status.passed.name: 0,
            Status.failed.name: 0,
            Status.skipped.name: 0,
            Status.untested.name: 0
        }
        self.feature_summary = summary_zero_data.copy()
        self.rule_summary = summary_zero_data.copy()
        self.scenario_summary = summary_zero_data.copy()
        self.step_summary = {Status.undefined.name: 0}
        self.step_summary.update(summary_zero_data)
        self.duration = 0.0
        self.run_starttime = 0
        self.run_endtime = 0
        self.failed_scenarios = []
        self.show_rules = True

    def testrun_started(self, timestamp=None):
        if timestamp is None:
            timestamp = time_now()
        self.run_starttime = timestamp

    def testrun_finished(self, timestamp=None):
        if timestamp is None:
            timestamp = time_now()
        self.run_endtime = timestamp

    def print_failing_scenarios(self, stream=None):
        if stream is None:
            stream = self.stream

        stream.write("\nFailing scenarios:\n")
        for scenario in self.failed_scenarios:
            stream.write(u"  %s  %s\n" % (scenario.location, scenario.name))

    def compute_summary_sums(self):
        """(Re)Compute summary sum of all counts (except: all)."""
        summaries = [
            self.feature_summary,
            self.rule_summary,
            self.scenario_summary,
            self.step_summary
        ]
        for summary in summaries:
            summary["all"] = compute_summary_sum(summary)

    def print_summary(self, stream=None, with_duration=True):
        if stream is None:
            stream = self.stream

        self.compute_summary_sums()
        has_rules = (self.rule_summary["all"] > 0)

        stream.write(format_summary("feature", self.feature_summary))
        if self.show_rules and has_rules:
            # -- HINT: Show only rules, if any exists.
            self.stream.write(format_summary("rule", self.rule_summary))
        stream.write(format_summary("scenario", self.scenario_summary))
        stream.write(format_summary("step", self.step_summary))

        # -- DURATION:
        if with_duration:
            timings = (int(self.duration / 60.0), self.duration % 60)
            stream.write('Took %dm%02.3fs\n' % timings)

    # -- REPORTER-API:
    def feature(self, feature):
        if self.run_starttime == 0:
            # -- DISCOVER: TEST-RUN started.
            self.testrun_started()

        self.process_feature(feature)

    def end(self):
        self.testrun_finished()

        # -- SHOW FAILED SCENARIOS (optional):
        if self.show_failed_scenarios and self.failed_scenarios:
            self.print_failing_scenarios()
            self.stream.write("\n")

        # -- SHOW SUMMARY COUNTS:
        self.print_summary()

    def process_run_items_for(self, parent):
        for run_item in parent:
            if isinstance(run_item, Rule):
                self.process_rule(run_item)
            elif isinstance(run_item, ScenarioOutline):
                self.process_scenario_outline(run_item)
            else:
                # assert isinstance(run_item, Scenario)
                self.process_scenario(run_item)

    def process_feature(self, feature):
        self.duration += feature.duration
        self.feature_summary[feature.status.name] += 1
        self.process_run_items_for(feature)

    def process_rule(self, rule):
        self.rule_summary[rule.status.name] += 1
        self.process_run_items_for(rule)

    def process_scenario(self, scenario):
        if scenario.status == Status.failed:
            self.failed_scenarios.append(scenario)

        self.scenario_summary[scenario.status.name] += 1
        for step in scenario:
            self.step_summary[step.status.name] += 1

    def process_scenario_outline(self, scenario_outline):
        for scenario in scenario_outline.scenarios:
            self.process_scenario(scenario)



# -*- coding: UTF-8 -*-
# pylint: disable=missing-docstring
"""
Extended tag-expression model that supports tag-matchers.

Provides model classes to evaluate parsed boolean tag expressions.

.. code-block:: python

    # -- Expression := a and b
    expression = And(Literal("a"), Literal("b"))
    assert True  == expression.evaluate(["a", "b"])
    assert False == expression.evaluate(["a"])
    assert False == expression.evaluate([])

    # -- Expression := a or b
    expression = Or(Literal("a"), Literal("b"))
    assert True  == expression.evaluate(["a", "b"])
    assert True  == expression.evaluate(["a"])
    assert False == expression.evaluate([])

    # -- Expression := not a
    expression = Not(Literal("a"))
    assert False == expression.evaluate(["a"])
    assert True  == expression.evaluate(["other"])
    assert True  == expression.evaluate([])

    # -- Expression := (a or b) and c
    expression = And(Or(Literal("a"), Literal("b")), Literal("c"))
    assert True  == expression.evaluate(["a", "c"])
    assert False == expression.evaluate(["c", "other"])
    assert False == expression.evaluate([])
"""

from __future__ import absolute_import
from fnmatch import fnmatchcase
import glob
from .model import Expression


# -----------------------------------------------------------------------------
# TAG-EXPRESSION MODEL CLASSES:
# -----------------------------------------------------------------------------
class Matcher(Expression):
    """Matches one or more similar tags by using wildcards.
    Supports simple filename-matching / globbing wildcards only.

    .. code-block:: python

        # -- CASE: Tag starts-with "foo."
        matcher1 = Matcher("foo.*")
        assert True == matcher1.evaluate(["foo.bar"])

        # -- CASE: Tag ends-with ".foo"
        matcher2 = Matcher("*.foo")
        assert True == matcher2.evaluate(["bar.foo"])
        assert True == matcher2.evaluate(["bar.baz_more.foo"])

        # -- CASE: Tag contains "foo"
        matcher3 = Matcher("*.foo.*")
        assert True == matcher3.evaluate(["bar.foo.more"])
        assert True == matcher3.evaluate(["bar.foo"])

    .. see:: :mod:`fnmatch`
    """
    # pylint: disable=too-few-public-methods
    def __init__(self, pattern):
        super(Matcher, self).__init__()
        self.pattern = pattern

    @property
    def name(self):
        return self.pattern

    def evaluate(self, values):
        for value in values:
            # -- REQUIRE: case-sensitive matching
            if fnmatchcase(value, self.pattern):
                return True
        # -- OTHERWISE: no-match
        return False

    def __str__(self):
        return self.pattern

    def __repr__(self):
        return "Matcher('%s')" % self.pattern

    @staticmethod
    def contains_wildcards(text):
        """Indicates if text contains supported wildcards."""
        # -- NOTE: :mod:`glob` wildcards are same as :mod:`fnmatch`
        return glob.has_magic(text)



# -*- coding: UTF-8 -*-
"""
Tag expression v2 parser, based on :mod:`cucumber-tag-expressions`.

Extension(s):

* Supports matching of tag names, like: "@foo.* or @*.foo"


.. seealso::

    * https://docs.cucumber.io
    * https://docs.cucumber.io/cucumber/api/#tag-expressions
"""

from __future__ import absolute_import
from cucumber_tag_expressions.parser import (
    TagExpressionParser as _TagExpressionParser,
    # PROVIDE: Similar interface like: cucumber_tag_expressions.parser
    TagExpressionError
)
from cucumber_tag_expressions.model import Literal
from .model_ext import Matcher


class TagExpressionParser(_TagExpressionParser):
    """Extended TagExpressionParser to parse boolean tag-expressions.
    Extended TagExpressionParser supports matching tag names by name-patterns.

    TagExpressionParser supports boolean operations:

    * and (as binary operation:  a and b)
    * or  (as binary operation:  a or b)
    * not (as unary operation:   not a)

    In addition, parenthesis can be used to group expressions, like::

        a and (b or c)
        (a and not b) or (c and d)

    EXAMPLES:

    .. code-block:: python

        # -- UNARY OPTIONS
        text11 = "not foo" = "(not foo)"
        expression = TagExpressionParser.parse(text11)
        assert False == expression.evaluate(["foo"])
        assert True  == expression.evaluate(["other"])

        # -- BINARY OPERATIONS:
        text21 = "foo and bar" = "(foo and bar)"
        expression = TagExpressionParser.parse(text21)
        assert True  == expression.evaluate(["foo", "bar"])
        assert False == expression.evaluate(["foo"])
        assert False == expression.evaluate([])

        text22 = "foo or bar"  = "(foo or bar)"
        expression = TagExpressionParser.parse(text22)
        assert True  == expression.evaluate(["foo", "bar"])
        assert True  == expression.evaluate(["foo", "other"])
        assert False == expression.evaluate([])

        # -- TAG MATCHING EXTENSION: Match tag-parts with wildcards
        text31 = "foo.* or bar"  = "(foo.* or bar)"
        expression = TagExpressionParser.parse(text31)
        assert True  == expression.evaluate(["foo.bar", "bar"])
        assert True  == expression.evaluate(["other", "foo.more"])
        assert False == expression.evaluate([])


    .. note::

        Name wildcard expressions are based on :mod:`fnmatch`
        (File name pattern matching).

    .. seealso::

        * :class:`cucumber_tag_expressions.parser:TagExpressionParser`
        * https://docs.cucumber.io
        * https://docs.cucumber.io/cucumber/api/#tag-expressions
    """

    @classmethod
    def make_operand(cls, text):
        """Creates operand-object from parsed text."""
        if Matcher.contains_wildcards(text):
            # -- USE MATCHER: Text contains matcher-wildcards.
            return Matcher(text)
        else:
            return Literal(text)



# -*- coding: utf-8 -*-
"""
Compatibility of :module:`collections` between different Python versions.
"""

from __future__ import absolute_import
import warnings
# pylint: disable=unused-import
try:
    # -- SINCE: Python2.7
    from collections import OrderedDict
except ImportError:     # pragma: no cover
    try:
        # -- BACK-PORTED FOR: Python 2.4 .. 2.6
        from ordereddict import OrderedDict
    except ImportError:
        message = "collections.OrderedDict is missing: Install 'ordereddict'."
        warnings.warn(message)
        # -- BACKWARD-COMPATIBLE: Better than nothing (for behave use case).
        OrderedDict = dict

try:
    from collections import UserDict
except ImportError:      # pragma: no cover
    class UserDict(object):
        """Emulate collections.UserDict class in python3."""
        def __init__(self, data=None):
            if data is None:
                data = {}
            self.data = data

        def __len__(self):
            return len(self.data)

        def __iter__(self):
            return len(self.data)

        def keys(self):
            return self.data.keys()

        def values(self):
            return self.data.values()

        def items(self):
            return self.data.items()



# -*- coding: utf-8 -*-

from __future__ import absolute_import, print_function
import codecs
import sys
import six
from behave.version import VERSION as BEHAVE_VERSION
from behave.configuration import Configuration
from behave.exception import ConstraintError, ConfigError, \
    FileNotFoundError, InvalidFileLocationError, InvalidFilenameError, \
    ModuleNotFoundError, ClassNotFoundError, InvalidClassError
from behave.parser import ParserError
from behave.runner import Runner
from behave.runner_util import print_undefined_step_snippets, reset_runtime
from behave.textutil import compute_words_maxsize, text as _text
from behave.runner_plugin import RunnerPlugin
# PREPARED: from behave.importer import make_scoped_class_name


# ---------------------------------------------------------------------------
# CONSTANTS:
# ---------------------------------------------------------------------------
DEBUG = __debug__
TAG_HELP = """
Scenarios inherit tags that are declared on the Feature level.
The simplest TAG_EXPRESSION is simply a tag::

    --tags=@dev

You may even leave off the "@" - behave doesn't mind.

You can also exclude all features / scenarios that have a tag,
by using boolean NOT::

    --tags="not @dev"

A tag expression can also use a logical OR::

    --tags="@dev or @wip"

The --tags option can be specified several times,
and this represents logical AND,
for instance this represents the boolean expression::

    --tags="(@foo or not @bar) and @zap"

You can also exclude several tags::

    --tags="not (@fixme or @buggy)"
""".strip()


# ---------------------------------------------------------------------------
# WORK-HORSE:
# ---------------------------------------------------------------------------
def run_behave(config, runner_class=None):
    """Run behave with configuration (and optional runner class).

    :param config:          Configuration object for behave.
    :param runner_class:    Runner class to use or none (use Runner class).
    :return:    0, if successful. Non-zero on failure.

    .. note:: BEST EFFORT, not intended for multi-threaded usage.
    """
    # pylint: disable=too-many-branches, too-many-statements, too-many-return-statements

    if config.version:
        print("behave " + BEHAVE_VERSION)
        return 0

    if config.tags_help:
        print(TAG_HELP)
        return 0

    if  config.lang == "help" or config.lang_list:
        print_language_list()
        return 0

    if config.lang_help:
        # -- PROVIDE HELP: For one, specific language
        language = config.lang_help
        print_language_help(language)
        return 0

    if not config.format:
        config.format = [config.default_format]
    elif config.format and "format" in config.defaults:
        # -- CASE: Formatter are specified in behave configuration file.
        #    Check if formatter are provided on command-line, too.
        if len(config.format) == len(config.defaults["format"]):
            # -- NO FORMATTER on command-line: Add default formatter.
            config.format.append(config.default_format)
    if "help" in config.format:
        print_formatters()
        return 0

    if len(config.outputs) > len(config.format):
        print("CONFIG-ERROR: More outfiles (%d) than formatters (%d)." % \
              (len(config.outputs), len(config.format)))
        return 1

    if config.runner == "help":
        print_runners(config.runner_aliases)
        return 0

    # -- MAIN PART:
    runner = None
    failed = True
    try:
        reset_runtime()
        runner = RunnerPlugin(runner_class).make_runner(config)
        # print("USING RUNNER: {0}".format(make_scoped_class_name(runner)))
        failed = runner.run()
    except ParserError as e:
        print(u"ParserError: %s" % e)
    except ConfigError as e:
        print(u"ConfigError: %s" % e)
    except FileNotFoundError as e:
        print(u"FileNotFoundError: %s" % e)
    except InvalidFileLocationError as e:
        print(u"InvalidFileLocationError: %s" % e)
    except InvalidFilenameError as e:
        print(u"InvalidFilenameError: %s" % e)
    except ModuleNotFoundError as e:
        print(u"ModuleNotFoundError: %s" % e)
    except ClassNotFoundError as e:
        print(u"ClassNotFoundError: %s" % e)
    except InvalidClassError as e:
        print(u"InvalidClassError: %s" % e)
    except ImportError as e:
        print(u"%s: %s" % (e.__class__.__name__, e))
        if DEBUG:
            raise
    except ConstraintError as e:
        print(u"ConstraintError: %s" % e)
    except Exception as e:
        # -- DIAGNOSTICS:
        text = _text(e)
        print(u"Exception %s: %s" % (e.__class__.__name__, text))
        raise

    if config.show_snippets and runner and runner.undefined_steps:
        print_undefined_step_snippets(runner.undefined_steps,
                                      colored=config.color)

    return_code = 0
    if failed:
        return_code = 1
    return return_code


# ---------------------------------------------------------------------------
# MAIN SUPPORT FOR: run_behave()
# ---------------------------------------------------------------------------
def print_language_list(file=None):
    """Print list of supported languages, like:

    * English
    * French
    * German
    * ...
    """
    from behave.i18n import languages

    print_ = lambda text: print(text, file=file)
    if six.PY2:
        # -- PYTHON2: Overcome implicit encode problems (encoding=ASCII).
        file = codecs.getwriter("UTF-8")(file or sys.stdout)

    iso_codes = languages.keys()
    print("AVAILABLE LANGUAGES:")
    for iso_code in sorted(iso_codes):
        native = languages[iso_code]["native"]
        name = languages[iso_code]["name"]
        print_(u"  %s: %s / %s" % (iso_code, native, name))


def print_language_help(language, file=None):
    from behave.i18n import languages
    # if stream is None:
    #     stream = sys.stdout
    #     if six.PY2:
    #         # -- PYTHON2: Overcome implicit encode problems (encoding=ASCII).
    #         stream = codecs.getwriter("UTF-8")(sys.stdout)

    print_ = lambda text: print(text, file=file)
    if six.PY2:
        # -- PYTHON2: Overcome implicit encode problems (encoding=ASCII).
        file = codecs.getwriter("UTF-8")(file or sys.stdout)

    if language not in languages:
        print_("%s is not a recognised language: try --lang-list" % language)
        return 1

    trans = languages[language]
    print_(u"Translations for %s / %s" % (trans["name"], trans["native"]))
    for kw in trans:
        if kw in "name native".split():
            continue
        print_(u"%16s: %s" % (kw.title().replace("_", " "),
                             u", ".join(w for w in trans[kw] if w != "*")))


def print_formatters(file=None):
    """Prints the list of available formatters and their description.

    :param file:  Optional, output file to use (default: sys.stdout).
    """
    from behave.formatter._registry  import format_items
    from operator import itemgetter

    print_ = lambda text: print(text, file=file)

    formatter_items = sorted(format_items(resolved=True), key=itemgetter(0))
    formatter_names = [item[0]  for item in formatter_items]
    column_size = compute_words_maxsize(formatter_names)
    schema = u"  %-"+ _text(column_size) +"s  %s"
    problematic_formatters = []

    print_("AVAILABLE FORMATTERS:")
    for name, formatter_class in formatter_items:
        formatter_description = getattr(formatter_class, "description", "")
        formatter_error = getattr(formatter_class, "error", None)
        if formatter_error:
            # -- DIAGNOSTICS: Indicate if formatter definition has a problem.
            problematic_formatters.append((name, formatter_error))
        else:
            # -- NORMAL CASE:
            print_(schema % (name, formatter_description))

    if problematic_formatters:
        print_("\nUNAVAILABLE FORMATTERS:")
        for name, formatter_error in problematic_formatters:
            print_(schema % (name, formatter_error))


def print_runners(runner_aliases, file=None):
    """Print a list of known test runner classes that can be used with the
    command-line option ``--runner=RUNNER_CLASS``.

    :param runner_aliases:  List of known runner aliases (as strings)
    :param file:  Optional, to redirect print-output to a file.
    """
    # MAYBE: file = file or sys.stdout
    print_ = lambda text: print(text, file=file)

    runner_names = sorted(runner_aliases.keys())
    column_size = compute_words_maxsize(runner_names)
    schema1 = u"  %-"+ _text(column_size) +"s  = %s%s"
    schema2 = u"  %-"+ _text(column_size) +"s    %s"
    problematic_runners = []

    print_("AVAILABLE RUNNERS:")
    for runner_name in runner_names:
        scoped_class_name = runner_aliases[runner_name]
        problem = RunnerPlugin.make_problem_description(scoped_class_name, use_details=True)
        if problem:
            problematic_runners.append((runner_name, problem))
        else:
            # -- NORMAL CASE:
            print_(schema1 % (runner_name, scoped_class_name, ""))

    if problematic_runners:
        print_("\nUNAVAILABLE RUNNERS:")
        for runner_name, problem_description in problematic_runners:
            print_(schema2 % (runner_name, problem_description))


# ---------------------------------------------------------------------------
# MAIN FUNCTIONS:
# ---------------------------------------------------------------------------
def main(args=None):
    """Main function to run behave (as program).

    :param args:    Command-line args (or string) to use.
    :return: 0, if successful. Non-zero, in case of errors/failures.
    """
    config = Configuration(args)
    return run_behave(config)


if __name__ == "__main__":
    # -- EXAMPLE: main("--version")
    sys.exit(main())



# -*- coding: UTF-8 -*-
"""
Supports some active-tags related Python features (similar to: feature flags).
"""

from __future__ import absolute_import
from behave.tag_matcher import BoolValueObject
from behave.python_feature import PythonFeature


# -----------------------------------------------------------------------------
# SUPPORTED: ACTIVE-TAGS
# -----------------------------------------------------------------------------
# -- PYTHON FEATURE, like: @use.with_python.feature.coroutine=yes
ACTIVE_TAG_VALUE_PROVIDER = {
    "python.feature.coroutine": BoolValueObject(PythonFeature.has_coroutine()),
    "python.feature.asyncio.coroutine_decorator":
        BoolValueObject(PythonFeature.has_asyncio_coroutine_decorator()),
    "python.feature.async_function": BoolValueObject(PythonFeature.has_async_function()),
    "python.feature.async_keyword": BoolValueObject(PythonFeature.has_async_keyword()),

    # -- DEPRECATING (older active-tag names):
    "python_has_coroutine": BoolValueObject(PythonFeature.has_coroutine()),
    "python_has_asyncio.coroutine_decorator":
        BoolValueObject(PythonFeature.has_asyncio_coroutine_decorator()),
    "python_has_async_function": BoolValueObject(PythonFeature.has_async_function()),
    "python_has_async_keyword": BoolValueObject(PythonFeature.has_async_keyword()),
}



# -*- coding: UTF-8 -*-
"""
Recipe for loading additional step definitions from sub-directories
in the "features/steps" directory.

.. code-block::

    # -- FILE: features/steps/use_substep_dirs.py
    # REQUIRES: path.py
    from behave.runner_util import load_step_modules
    from path import Path

    HERE = Path(__file__).dirname()
    SUBSTEP_DIRS = list(HERE.walkdirs())
    load_step_modules(SUBSTEP_DIRS)
"""



# -*- coding: UTF-8 -*-
"""
Specifies the common interface for runner(s)/runner-plugin(s).

.. seealso::

    * https://pymotw.com/3/abc/index.html

"""

from __future__ import absolute_import
from abc import ABCMeta, abstractmethod
from sys import version_info as _version_info
from six import add_metaclass


_PYTHON_VERSION = _version_info[:2]


@add_metaclass(ABCMeta)
class ITestRunner(object):
    """Interface that a test runner-class should provide:

    * Constructor: with config parameter object (at least) and some kw-args.
    * run() method without any args.
    """

    @abstractmethod
    def __init__(self, config, **kwargs):
        self.config = config
        # MAYBE: self.undefined_steps = []

    @abstractmethod
    def run(self):
        """Run the selected features.

        :return: True, if test-run failed. False, on success.
        :rtype: bool
        """
        return False

    # if _PYTHON_VERSION < (3, 3):
    #     # -- HINT: @abstractproperty, deprecated since Python 3.3
    #     from abc import abstractproperty
    #     @abstractproperty
    #     def undefined_steps(self):
    #         raise NotImplementedError()
    # else:
    @property
    @abstractmethod
    def undefined_steps(self):
        """Provides list of undefined steps that were found during the test-run.

        :return: List of unmatched steps (as string) in feature-files.
        """
        raise NotImplementedError()
        # return NotImplemented



# -*- coding: UTF-8 -*-
# pylint: disable=line-too-long
"""
Gherkin parser with `Gherkin v6 grammar`_ (and Gherkin v5 or older).

Gherkin v6 grammar extensions:

* Rule keyword added
* Aliases for Scenario, ScenarioOutline to better correspond to `Example Mapping`_

A Rule (or: business rule) allows to group multiple Scenarios::

    # -- RULE GRAMMAR PSEUDO-CODE:
    @tag1 @tag2
    Rule: Optional Rule Title...
        Description?        #< CARDINALITY: 0..1 (optional)
        Background?         #< CARDINALITY: 0..1 (optional)
        Scenario*           #< CARDINALITY: 0..N (many)
        ScenarioOutline*    #< CARDINALITY: 0..N (many)

Keyword aliases:

    | Concept          | Gherkin v6         | Alias (Gherkin v5) |
    | Scenario         | Example            | Scenario           |
    | Scenario Outline | Scenario Template  | Scenario Outline   |

.. seealso::

    * `Gherkin v6 grammar`_

    EXAMPLE MAPPING:

    * Cucumber: `Example Mapping`_
    * Cucumber: `Example Mapping Webinar`_
    * https://docs.cucumber.io/bdd/example-mapping/

.. _`Gherkin v6 grammar`: https://github.com/cucumber/cucumber/blob/master/gherkin/gherkin.berp
.. _`Example Mapping`: https://cucumber.io/blog/example-mapping-introduction/
.. _`Example Mapping Webinar`: https://cucumber.io/blog/example-mapping-webinar/
"""
# pylint: enable=line-too-long

from __future__ import absolute_import, with_statement
import logging
import re
import sys
import six
from behave import model, i18n
from behave.textutil import text as _text


DEFAULT_LANGUAGE = "en"


def parse_file(filename, language=None):
    with open(filename, "rb") as f:
        # file encoding is assumed to be utf8. Oh, yes.
        data = f.read().decode("utf8")
    return parse_feature(data, language, filename)


def parse_feature(data, language=None, filename=None):
    # ALL data operated on by the parser MUST be unicode
    assert isinstance(data, six.text_type)

    try:
        result = Parser(language).parse(data, filename)
    except ParserError as e:
        e.filename = filename
        raise

    return result


def parse_rule(text, language=None, filename=None):
    """Parse a rule with its background and scenario(s).

    :param text: Multi-line text with rule to parse (as unicode).
    :param language:  i18n language identifier (optional).
    :param filename:  Filename (optional).
    :return: Parsed steps (if successful).
    """
    assert isinstance(text, six.text_type)
    try:
        result = Parser(language, variant="rule").parse_rule(text, filename)
    except ParserError as e:
        e.filename = filename
        raise
    return result

def parse_steps(text, language=None, filename=None):
    """Parse a number of steps a multi-line text from a scenario.
    Scenario line with title and keyword is not provided.

    :param text: Multi-line text with steps to parse (as unicode).
    :param language:  i18n language identifier (optional).
    :param filename:  Filename (optional).
    :return: Parsed steps (if successful).
    """
    assert isinstance(text, six.text_type)
    try:
        result = Parser(language, variant="steps").parse_steps(text, filename)
    except ParserError as e:
        e.filename = filename
        raise
    return result


def parse_step(text, language=None, filename=None):
    """Parse one step as multi-line text.

    :param text: Multi-line text with step to parse (as unicode).
    :param language:  i18n language identifier (optional).
    :param filename:  Filename (optional).
    :return: Parsed step (if successful).
    """
    steps = parse_steps(text, language=language, filename=filename)
    assert len(steps) == 1
    return steps[0]


def parse_tags(text):
    """
    Parse tags from text (one or more lines, as string).

    :param text: Multi-line text with tags to parse (as unicode).
    :return: List of tags (if successful).
    """
    # assert isinstance(text, unicode)
    if not text:
        return []
    return Parser(variant="tags").parse_tags(text)


class ParserError(Exception):
    @staticmethod
    def make_annotated(message, line_number, line_text=None, reason=None):
        """Make annotated message enriched w/ line_number, line_text."""
        if line_number:
            message += u" at line %d" % line_number
        if line_text:
            message += u': "%s"' % line_text.strip()
        if reason:
            message += u"\nREASON: %s" % reason
        return message

    def __init__(self, message, line, filename=None, line_text=None,
                 reason=None, use_annotated_message=True):
        if use_annotated_message:
            message = self.make_annotated(message, line, line_text, reason)

        super(ParserError, self).__init__(message)
        self.line = line    # Line number of parse failure.
        self.line_text = line_text
        self.filename = filename

    def __str__(self):
        arg0 = _text(self.args[0])
        if self.filename:
            filename = _text(self.filename, sys.getfilesystemencoding())
            return u'Failed to parse "%s": %s' % (filename, arg0)
        # -- OTHERWISE:
        return u"Failed to parse <string>: %s" % arg0

    if six.PY2:
        __unicode__ = __str__
        __str__ = lambda self: self.__unicode__().encode("utf-8")


class Parser(object):
    """Feature file parser for behave."""
    # pylint: disable=too-many-instance-attributes

    def __init__(self, language=None, variant=None):
        if not variant:
            variant = "feature"
        self.language = language
        self.variant = variant
        self.state = "init"
        self.line = 0
        self.last_step_type = None
        self.multiline_start = None
        self.multiline_leading = None
        self.multiline_terminator = None

        self.filename = None
        self.scenario_container = None  # Feature or Rule.
        self.feature = None
        self.rule = None
        self.parent = None
        self.statement = None
        self.tags = []
        self.lines = []
        self.table = None
        self.examples = None
        self.keywords = None
        if self.language:
            self.keywords = i18n.languages[self.language]
        # NOT-NEEDED: self.reset()

    def reset(self):
        # This can probably go away.
        if self.language:
            self.keywords = i18n.languages[self.language]
        else:
            self.keywords = None

        self.state = "init"
        self.line = 0
        self.last_step_type = None
        self.multiline_start = None
        self.multiline_leading = None
        self.multiline_terminator = None

        self.filename = None
        self.scenario_container = None  # Feature or Rule.
        self.feature = None
        self.rule = None
        self.parent = None
        self.statement = None
        self.tags = []
        self.lines = []
        self.table = None
        self.examples = None

    def parse(self, text, filename=None):
        self.reset()
        self.filename = filename

        for line in text.split("\n"):
            self.line += 1
            if not line.strip() and self.state != "multiline_text":
                # -- SKIP EMPTY LINES, except in multiline string args.
                continue
            self.action(line)

        if self.table:
            self.action_table("")

        feature = self.feature
        if feature:
            feature.parser = self
        self.reset()
        return feature

    def _build_feature(self, keyword, line):
        name = line[len(keyword) + 1:].strip()
        language = self.language or DEFAULT_LANGUAGE
        feature = model.Feature(self.filename, self.line, keyword, name,
                                tags=self.tags, language=language)
        self.feature = feature
        self.scenario_container = feature
        self.rule = None
        # -- RESET STATE:
        self.tags = []

    def _build_rule_statement(self, keyword, line):
        name = line[len(keyword) + 1:].strip()
        rule = model.Rule(self.filename, self.line, keyword, name,
                          tags=self.tags)
        self.rule = rule
        self.scenario_container = rule
        self.statement = rule
        self.feature.add_rule(self.statement)
        # -- RESET STATE:
        self.tags = []

    def _build_background_statement(self, keyword, line):
        if self.tags:
            msg = u"Background supports no tags: @%s" % (u" @".join(self.tags))
            raise ParserError(msg, self.line, self.filename, line)
        elif self.scenario_container and self.scenario_container.background:
            if self.scenario_container.background.steps:
                # -- HINT: Rule may have default background w/o steps.
                msg = u"Second Background (can have only one)"
                raise ParserError(msg, self.line, self.filename, line)
        name = line[len(keyword) + 1:].strip()
        background = model.Background(self.filename, self.line, keyword, name)
        self.scenario_container.add_background(background)
        self.statement = background

    def _build_scenario_statement(self, keyword, line):
        name = line[len(keyword) + 1:].strip()
        scenario = model.Scenario(self.filename, self.line, keyword, name,
                                  tags=self.tags)
        self.statement = scenario
        self.scenario_container.add_scenario(scenario)
        # OLD: self.feature.add_scenario(self.statement)
        # -- RESET STATE:
        self.tags = []

    def _build_scenario_outline_statement(self, keyword, line):
        # pylint: disable=C0103
        #   C0103   Invalid name "build_scenario_outline_statement", too long.
        name = line[len(keyword) + 1:].strip()
        template = model.ScenarioOutline(self.filename, self.line, keyword, name,
                                         tags=self.tags)
        self.statement = template
        self.scenario_container.add_scenario(template)
        # OLD: self.feature.add_scenario(self.statement)
        # -- RESET STATE:
        self.tags = []

    def _build_examples(self, keyword, line):
        if not isinstance(self.statement, model.ScenarioOutline):
            message = u"Examples must only appear inside scenario outline"
            raise ParserError(message, self.line, self.filename, line)
        name = line[len(keyword) + 1:].strip()
        self.examples = model.Examples(self.filename, self.line,
                                       keyword, name, tags=self.tags)
        # pylint: disable=E1103
        #   E1103   Instance of "Background" has no "examples" member
        #           (but some types could not be inferred).
        self.statement.examples.append(self.examples)

        # -- RESET STATE:
        self.tags = []


    def diagnose_feature_usage_error(self):
        if self.feature:
            return "Multiple features in one file are not supported."
        # -- OTHERWISE:
        return "Feature should not be used here."

    def diagnose_rule_usage_error(self):
        # pylint: disable=no-self-use
        return "Rule should not be used here."

    def diagnose_background_usage_error(self):
        if self.scenario_container and self.scenario_container.scenarios:
            # -- CASE: Feature or Rule
            return "Background may not occur after Scenario/ScenarioOutline."
        elif self.tags:
            return "Background does not support tags."
        # -- OTHERWISE:
        return "Background should not be used here."

    def diagnose_scenario_usage_error(self):
        if not self.scenario_container:
            return "Scenario may not occur before Feature."
        # -- OTHERWISE:
        return "Scenario should not be used here."

    def diagnose_scenario_outline_usage_error(self): # pylint: disable=invalid-name
        if not self.scenario_container:
            return "ScenarioOutline may not occur before Feature."
        # -- OTHERWISE:
        return "ScenarioOutline should not be used here."

    def ask_parse_failure_oracle(self, line):
        """
        Try to find the failure reason when a parse failure occurs:

            Oracle, oracle, ... what went wrong?

        :param line:  Text line where parse failure occured (as string).
        :return: Reason (as string) if an explanation is found.
                 Otherwise, empty string or None.
        """
        # pylint: disable=too-many-return-statements
        feature_keyword = self.match_keyword("feature", line)
        if feature_keyword:
            return self.diagnose_feature_usage_error()
        rule_keyword = self.match_keyword("rule", line)
        if rule_keyword:
            return self.diagnose_rule_usage_error()
        background_keyword = self.match_keyword("background", line)
        if background_keyword:
            return self.diagnose_background_usage_error()
        scenario_keyword = self.match_keyword("scenario", line)
        if scenario_keyword:
            return self.diagnose_scenario_usage_error()
        scenario_outline_keyword = self.match_keyword("scenario_outline", line)
        if scenario_outline_keyword:
            return self.diagnose_scenario_outline_usage_error()
        # -- OTHERWISE:
        if self.variant == "feature" and not self.feature:
            return "No feature found."
        # -- FINALLY: No glue what went wrong.
        return None

    def action(self, line):
        if line.strip().startswith("#") and self.state != "multiline_text":
            if self.state != "init" or self.tags or self.variant != "feature":
                return

            # -- DETECT: language comment (at begin of feature file; state=init)
            line = line.strip()[1:].strip()
            if line.lstrip().lower().startswith("language:"):
                language = line[9:].strip()
                self.language = language
                self.keywords = i18n.languages[language]
            return

        func = getattr(self, "action_" + self.state, None)
        if func is None:
            line = line.strip()
            msg = u"Parser in unknown state %s;" % self.state
            raise ParserError(msg, self.line, self.filename, line)

        if not func(line):
            line = line.strip()
            msg = u'\nParser failure in state=%s' % self.state
            reason = self.ask_parse_failure_oracle(line)
            raise ParserError(msg, self.line, self.filename,
                              line_text=line, reason=reason)

    def action_init(self, line):
        line = line.strip()
        if line.startswith("@"):
            self.tags.extend(self.parse_tags(line))
            return True

        feature_keyword = self.match_keyword("feature", line)
        if feature_keyword:
            self._build_feature(feature_keyword, line)
            self.state = "feature"
            return True
        return False

    # pylint: disable=invalid-name
    def subaction_detect_taggable_statement(self, line):
        """Subaction is used after first tag line is detected.
        Additional lines with tags or taggable_statement follow.

        Taggable statements (excluding Feature) are:
        * Rule
        * Scenario
        * ScenarioOutline
        * Examples (within ScenarioOutline)
        """
        if line.startswith("@"):
            self.tags.extend(self.parse_tags(line))
            self.state = "taggable_statement"
            return True

        rule_keyword = self.match_keyword("rule", line)
        if rule_keyword:
            # MAYBE: Finish last rule statement
            self._build_rule_statement(rule_keyword, line)
            self.state = "rule"
            return True

        scenario_keyword = self.match_keyword("scenario", line)
        if scenario_keyword:
            self._build_scenario_statement(scenario_keyword, line)
            self.state = "scenario"
            return True

        template_keyword = self.match_keyword("scenario_outline", line)
        if template_keyword:
            self._build_scenario_outline_statement(template_keyword, line)
            self.state = "scenario"
            return True

        examples_keyword = self.match_keyword("examples", line)
        if examples_keyword:
            self._build_examples(examples_keyword, line)
            self.state = "table"
            return True

        # -- OTHERWISE:
        return False
    # pylint: enable=invalid-name

    def action_feature(self, line):
        line = line.strip()
        # OLD: if self.subaction_detect_next_scenario(line):
        if self.subaction_detect_taggable_statement(line):
            # -- DETECTED: Next Rule, Scenario, ScenarioOutline (or tags)
            return True

        background_keyword = self.match_keyword("background", line)
        if background_keyword:
            self._build_background_statement(background_keyword, line)
            self.state = "background"
            return True

        self.feature.description.append(line)
        return True

    def action_taggable_statement(self, line):
        """Entered after first tag for Scenario/ScenarioOutline or
        Examples is detected (= taggable_statement except Feature).

        Taggable statements (excluding Feature) are:
          * Scenario
          * ScenarioOutline
          * Examples (within ScenarioOutline)
        """
        line = line.strip()
        if self.subaction_detect_taggable_statement(line):
            # -- DETECTED: Next Scenario, ScenarioOutline or Examples (or tags)
            return True

        return False

    def action_background(self, line):
        """Entered when Background keyword/line is detected.
        Hunts/collects background description lines.

        DETECT:

        * first step of Background
        * next Scenario/ScenarioOutline.
        * any description line after Background keyword
        """
        # -- SAME AS: action_scenario(), only Background is used as self.statement.
        # REUSE: Already existing action.
        return self.action_scenario(line)

    def action_rule(self, line):
        """Entered when Rule keyword/line is detected.
        Hunts/collects rule description lines.

        DETECT:

        * any description line after Rule keyword (optional)
        * Backgroup statement (optional)
        * Scenario/ScenarioOutline statements (many)
        """
        # -- SIMILAR TO: action_feature()
        line = line.strip()
        if self.subaction_detect_taggable_statement(line):
            # -- DETECTED: Next Rule, Scenario, ScenarioOutline (or tags)
            return True

        background_keyword = self.match_keyword("background", line)
        if background_keyword:
            self._build_background_statement(background_keyword, line)
            self.state = "background"
            return True

        self.rule.description.append(line)
        return True


    def action_scenario(self, line):
        """Entered when Scenario/ScenarioOutline keyword/line is detected.
        Hunts/collects scenario description lines.

        DETECT:

        * any description line after Scenario/ScenarioOutline keyword
        * first step of Scenario/ScenarioOutline
        * next Scenario/ScenarioOutline
        """
        self.last_step_type = None
        line = line.strip()
        step = self.parse_step(line)
        if step:
            # -- FIRST STEP DETECTED: End collection of description-part.
            self.state = "steps"
            self.statement.steps.append(step)
            return True

        # -- CASE: Detect next Scenario/ScenarioOutline
        #   * Background/Scenario with description, but without steps.
        #   * Title-only Background/Scenario without description and steps.
        if self.subaction_detect_taggable_statement(line):
            # -- DETECTED: Next Scenario, ScenarioOutline (or tags)
            return True

        # -- OTHERWISE: Add description line.
        # pylint: disable=E1103
        #   E1103   Instance of "Background" has no "description" member...
        self.statement.description.append(line)
        return True

    def action_steps(self, line):
        """
        Entered when first step is detected (or nested step parsing).

        Subcases:
          * step
          * multi-line text (doc-string), following a step
          * table, following a step
          * examples for a ScenarioOutline, after ScenarioOutline steps

        DETECT:
          * next Scenario/ScenarioOutline or Examples (in a ScenarioOutline)
        """
        # pylint: disable=R0911
        #   R0911   Too many return statements (8/6)
        stripped = line.lstrip()
        # if self.statement.steps:
        #     # -- ENSURE: Multi-line text follows a step.
        #     if stripped.startswith('"""') or stripped.startswith("'''"):
        #         # -- CASE: Multi-line text (docstring) after a step detected.
        #         self.state = "multiline_text"
        #         self.multiline_start = self.line
        #         self.multiline_terminator = stripped[:3]
        #         self.multiline_leading = line.index(stripped[0])
        #         return True

        if stripped.startswith('"""') or stripped.startswith("'''"):
            # -- CASE: Multi-line text (docstring) after a step detected.
            # REQUIRE: Multi-line text follows a step.
            if not self.statement.steps:
                raise ParserError("Multi-line text before any step",
                                  self.line, self.filename)

            self.state = "multiline_text"
            self.multiline_start = self.line
            self.multiline_terminator = stripped[:3]
            self.multiline_leading = line.index(stripped[0])
            return True

        line = line.strip()
        step = self.parse_step(line)
        if step:
            self.statement.steps.append(step)
            return True

        if self.subaction_detect_taggable_statement(line):
            # -- DETECTED: Next Scenario, ScenarioOutline or Examples (or tags)
            return True

        if line.startswith("|"):
            # -- CASE: TABLE-START detected for data-table of a step
            # OLD: assert self.statement.steps, "TABLE-START without step detected"
            if not self.statement.steps:
                raise ParserError("TABLE-START without step detected",
                                  self.line, self.filename)
            self.state = "table"
            return self.action_table(line)

        return False

    def action_multiline_text(self, line):
        """Parse remaining multi-line/docstring text below a step
        after the triple-quotes were detected:

        * triple-double-quotes or
        * triple-single-quotes

        Leading and trailing triple-quotes must be the same.

        :param line:  Parsed line, as part of a multi-line text (as string).
        """
        if line.strip().startswith(self.multiline_terminator):
            # -- CASE: Handle the end of a multi-line text part.
            # Store the multi-line text in the step object (and continue).
            this_step = self.statement.steps[-1]
            text = u"\n".join(self.lines)
            this_step.text = model.Text(text, u"text/plain", self.multiline_start)
            if this_step.name.endswith(":"):
                this_step.name = this_step.name[:-1]

            # -- RESET INTERNALS: For next step
            self.lines = []
            self.multiline_terminator = None
            self.state = "steps"    # NEXT-STATE: Accept additional step(s).
            return True

        # -- SPECIAL CASE: Strip trailing whitespace (whitespace normalization).
        # HINT: Required for Windows line-endings, like "\r\n", etc.
        text_line = line[self.multiline_leading:].rstrip()
        self.lines.append(text_line)

        # -- BETTER DIAGNOSTICS: May remove non-whitespace in execute_steps()
        removed_line_prefix = line[:self.multiline_leading]
        if removed_line_prefix.strip():
            message = u"BAD-INDENT in multiline text: "
            message += u"Line '%s' would strip leading '%s'" % \
                        (line, removed_line_prefix)
            raise ParserError(message, self.line, self.filename)
        return True

    def action_table(self, line):
        """Parse a table, with pipe-separated columns:

        * Data table of a step (after the step line)
        * Examples table of a ScenarioOutline
        """
        line = line.strip()
        if not line.startswith("|"):
            # -- CASE: End-of-table detected
            if self.examples:
                # -- CASE: Examples table of a ScenarioOutline
                self.examples.table = self.table
                self.examples = None
            else:
                # -- CASE: Data table of a step
                step = self.statement.steps[-1]
                step.table = self.table
                if step.name.endswith(":"):
                    step.name = step.name[:-1]

            # -- RESET: Parameters for parsing the next step(s).
            self.table = None
            self.state = "steps"
            return self.action_steps(line)

        if not re.match(r"^(|.+)\|$", line):
            logger = logging.getLogger("behave")
            logger.warning(u"Malformed table row at %s: line %i",
                           self.feature.filename, self.line)

        # -- SUPPORT: Escaped-pipe(s) in Gherkin cell values.
        #    Search for pipe(s) that are not preceeded with an escape char.
        cells = [cell.replace("\\|", "|").strip()
                 for cell in re.split(r"(?<!\\)\|", line[1:-1])]
        if self.table is None:
            # -- CASE: First row of the table
            self.table = model.Table(cells, self.line)
        else:
            # -- CASE: Following rows of the table
            if len(cells) != len(self.table.headings):
                raise ParserError(u"Malformed table", self.line, self.filename)
            self.table.add_row(cells, self.line)
        return True

    def match_keyword(self, keyword, line):
        if not self.keywords:
            self.language = DEFAULT_LANGUAGE
            self.keywords = i18n.languages[DEFAULT_LANGUAGE]
        for alias in self.keywords[keyword]:
            if line.startswith(alias + ":"):
                return alias
        return False

    def parse_rule(self, text, filename=None):
        """Parse rule with optional background and scenario(s).

        :param text:  Text that contains 0..* steps
        :return: List of parsed rule (as :class:`~behave.model:Rule` object).
        """
        assert isinstance(text, six.text_type)
        if not self.language:
            self.language = DEFAULT_LANGUAGE
        self.reset()
        self.filename = filename
        self.rule = model.Rule(filename, 0, u"rule", u"")
        self.statement = self.rule
        self.state = "rule"

        for line in text.split("\n"):
            self.line += 1
            if not line.strip() and self.state != "multiline_text":
                # -- SKIP EMPTY LINES, except in multiline string args.
                continue
            self.action(line)

        # -- FINALLY:
        if self.table:
            self.action_table("")
        steps = self.statement.steps
        return steps

    def parse_tags(self, line):
        """Parse a line with one or more tags:

          * A tag starts with the AT sign.
          * A tag consists of one word without whitespace chars.
          * Multiple tags are separated with whitespace chars
          * End-of-line comment is stripped.

        :param line:   Line with one/more tags to process.
        :raise ParserError: If syntax error is detected.
        """
        assert line.startswith("@")
        tags = []
        for word in line.split():
            if word.startswith("@"):
                tags.append(model.Tag(word[1:], self.line))
            elif word.startswith("#"):
                break   # -- COMMENT: Skip rest of line.
            else:
                # -- BAD-TAG: Abort here.
                message = u"tag: %s (line: %s)" % (word, line)
                raise ParserError(message, self.line, self.filename)
        return tags

    def parse_step(self, line):
        for step_type in ("given", "when", "then", "and", "but"):
            for kw in self.keywords[step_type]:
                # try to match the keyword; also attempt a purely lowercase
                # match if that'll work
                if not (line.startswith(kw) or
                        line.lower().startswith(kw.lower())):
                    # -- CASE: Line does not start w/ a step-keyword.
                    continue
                # -- HINT: Trailing SPACE is used for most keywords.
                # BUT: Keywords in some languages (like Chinese, Japanese, ...)
                #      do not need a whitespace as word separator.
                step_text_after_keyword = line[len(kw):].strip()
                if kw.startswith("*") and self.last_step_type:
                    # -- CASE: Generic steps and Given/When/Then steps are mixed.
                    # HINT: Inherit step type from last step.
                    step_type = self.last_step_type
                elif step_type in ("and", "but"):
                    if not self.last_step_type:
                        raise ParserError(u"No previous step",
                                          self.line, self.filename)
                    step_type = self.last_step_type
                else:
                    self.last_step_type = step_type

                keyword = kw.rstrip()  # HINT: Strip optional trailing SPACE.
                step = model.Step(self.filename, self.line,
                                  keyword, step_type, step_text_after_keyword)
                return step
        return None

    def parse_steps(self, text, filename=None):
        """Parse support for execute_steps() functionality that
        supports step with:

        * multiline text
        * table

        :param text:  Text that contains 0..* steps
        :return: List of parsed steps (as model.Step objects).
        """
        assert isinstance(text, six.text_type)
        if not self.language:
            self.language = DEFAULT_LANGUAGE
        self.reset()
        self.filename = filename
        self.statement = model.Scenario(filename, 0, u"scenario", u"")
        self.state = "steps"

        for line in text.split("\n"):
            self.line += 1
            if not line.strip() and self.state != "multiline_text":
                # -- SKIP EMPTY LINES, except in multiline string args.
                continue
            self.action(line)

        # -- FINALLY:
        if self.table:
            self.action_table("")
        steps = self.statement.steps
        return steps



# -*- coding: UTF-8 -*-
"""
Simplifies to specify runtime constraints in

* features/environment.py file
* features/steps/*.py" files
"""

from __future__ import absolute_import
import six
import sys
from behave.exception import ConstraintError


# ---------------------------------------------------------------------------
# UTILITY FUNCTIONS:
# ---------------------------------------------------------------------------
def require_min_python_version(minimal_version):
    """Simplifies to specify the minimal python version that is required.

    :param minimal_version: Minimum version (as string, tuple)
    :raises: behave.exception.ConstraintError
    """
    python_version = sys.version_info
    if isinstance(minimal_version, six.string_types):
        python_version = float("%s.%s" % sys.version_info[:2])
        minimal_version = float(minimal_version)
    elif not isinstance(minimal_version, tuple):
        raise TypeError("string or tuple (was: %s)" % type(minimal_version))

    if python_version < minimal_version:
        raise ConstraintError("python >= %s expected (was: %s)" % \
                              (minimal_version, python_version))


def require_min_behave_version(minimal_version):
    """Simplifies to specify the minimal behave version that is required.

    :param minimal_version: Minimum version (as string, tuple)
    :raises: behave.exception.ConstraintError
    """
    # -- SIMPLISTIC IMPLEMENTATION:
    from behave.version import VERSION as behave_version
    behave_version2 = behave_version.split(".")
    minimal_version2 = minimal_version.split(".")
    if behave_version2 < minimal_version2:
        # -- USE: Tuple comparison as version comparison.
        raise ConstraintError("behave >= %s expected (was: %s)" % \
                              (minimal_version, behave_version))



# -*- coding: UTF-8 -*-
"""
Common module for tag-expressions:

* v1: old tag expressions (deprecating; superceeded by: cucumber-tag-expressions)
* v2: cucumber-tag-expressions

.. seealso::

    * https://docs.cucumber.io
    * https://docs.cucumber.io/cucumber/api/#tag-expressions
"""

from __future__ import absolute_import
import six
# -- NEW CUCUMBER TAG-EXPRESSIONS (v2):
from .parser import TagExpressionParser
# -- OLD-STYLE TAG-EXPRESSIONS (v1):
# HINT: BACKWARD-COMPATIBLE (deprecating)
from .v1 import TagExpression


# -----------------------------------------------------------------------------
# FUNCTIONS:
# -----------------------------------------------------------------------------
def make_tag_expression(tag_expression_text):
    """Build a TagExpression object by parsing the tag-expression (as text).

    :param tag_expression_text:     Tag expression text to parse (as string).
    :return: TagExpression object to use.
    """
    parse_tag_expression = select_tag_expression_parser(tag_expression_text)
    return parse_tag_expression(tag_expression_text)


def parse_tag_expression_v1(tag_expression_parts):
    """Parse old style tag-expressions and build a TagExpression object."""
    # -- HINT: DEPRECATING
    if isinstance(tag_expression_parts, six.string_types):
        tag_expression_parts = tag_expression_parts.split()
    # print("parse_tag_expression_v1: %s" % " ".join(tag_expression_parts))
    return TagExpression(tag_expression_parts)


def parse_tag_expression_v2(tag_expression_text):
    """Parse cucumber-tag-expressions and build a TagExpression object."""
    text = tag_expression_text
    if not isinstance(text, six.string_types):
        # -- ASSUME: List of strings
        assert isinstance(text, (list, tuple))
        text = " and ".join(text)

    if "@" in text:
        # -- NORMALIZE: tag-expression text => Remove '@' tag decorators.
        text = text.replace("@", "")
    text = text.replace("  ", " ")
    # print("parse_tag_expression_v2: %s" % text)
    return TagExpressionParser.parse(text)


def check_for_complete_keywords(words, keywords):
    for keyword in keywords:
        for word in words:
            if keyword == word:
                return True
    return False


def select_tag_expression_parser(tag_expression_text):
    """Select/Auto-detect which version of tag-expressions is used.

    :param tag_expression_text:  Tag expression text (as string)
    :return: TagExpression parser to use (as function).
    """
    TAG_EXPRESSION_V1_KEYWORDS = [
        "~", "-", ","
    ]
    TAG_EXPRESSION_V2_KEYWORDS = [
        "and", "or", "not", "(", ")"
    ]

    text = tag_expression_text
    if not isinstance(text, six.string_types):
        # -- ASSUME: List of strings
        assert isinstance(text, (list, tuple))
        text = " ".join(text)

    text = text.replace("(", " ( ").replace(")", " ) ")
    words = text.split()
    contains_v1_keywords = any([(k in text) for k in TAG_EXPRESSION_V1_KEYWORDS])
    contains_v2_keywords = check_for_complete_keywords(words, TAG_EXPRESSION_V2_KEYWORDS)
    # contains_v2_keywords = any([(k in text) for k in TAG_EXPRESSION_V2_KEYWORDS])
    # DIAG: print("XXX select_tag_expression_parser: v1=%r, v2=%r, words.size=%d (tags: %r)" % \
    # DIAG:      (contains_v1_keywords, contains_v2_keywords, len(words), text))
    if contains_v2_keywords:
        # -- USE: Use cucumber-tag-expressions
        return parse_tag_expression_v2
    elif contains_v1_keywords or len(words) > 1:
        # -- CASE 1: "-@foo", "~@foo" (negated)
        # -- CASE 2: "@foo @bar"
        return parse_tag_expression_v1

    # -- OTHERWISSE: Use cucumber-tag-expressions
    # CASE: "@foo" (1 tag)
    return parse_tag_expression_v2



# -*- coding: UTF-8 -*-
"""
Provides a knowledge database if some Python features are supported
in the current python version.
"""

from __future__ import absolute_import
import sys


# -----------------------------------------------------------------------------
# CONSTANTS:
# -----------------------------------------------------------------------------
PYTHON_VERSION = sys.version_info[:2]


# -----------------------------------------------------------------------------
# CLASSES:
# -----------------------------------------------------------------------------
class PythonFeature(object):

    @staticmethod
    def has_asyncio_coroutine_decorator():
        """Indicates if python supports ``@asyncio.coroutine`` decorator.

        EXAMPLE::

            import asyncio
            @asyncio.coroutine
            def async_waits_seconds(duration):
                yield from asyncio.sleep(duration)

        :returns: True, if this python version supports this feature.

        .. since:: Python >= 3.4
        .. deprecated:: Since Python 3.8 (use async-function instead)
        """
        # -- NOTE: @asyncio.coroutine is deprecated in py3.8, removed in py3.10
        return (3, 4) <= PYTHON_VERSION < (3, 10)

    @staticmethod
    def has_async_function():
        """Indicates if python supports async-functions / async-keyword.

        EXAMPLE::

            import asyncio
            async def async_waits_seconds(duration):
                yield from asyncio.sleep(duration)

        :returns: True, if this python version supports this feature.
        .. since:: Python >= 3.5
        """
        return (3, 5) <= PYTHON_VERSION

    @classmethod
    def has_async_keyword(cls):
        return cls.has_async_function()

    @classmethod
    def has_coroutine(cls):
        return cls.has_async_function() or cls.has_asyncio_coroutine_decorator()



# -*- coding: UTF-8 -*-
# pylint: disable=redefined-builtin,unused-import
"""
Behave exception classes.

.. versionadded:: 1.2.7
"""

from __future__ import absolute_import
# -- USE MODERN EXCEPTION CLASSES:
# COMPATIBILITY: Emulated if not supported yet by Python version.
from behave.compat.exceptions import FileNotFoundError, ModuleNotFoundError


# ---------------------------------------------------------------------------
# EXCEPTION/ERROR CLASSES:
# ---------------------------------------------------------------------------
class ConstraintError(RuntimeError):
    """Used if a constraint/precondition is not fulfilled at runtime.

    .. versionadded:: 1.2.7
    """


class ConfigError(Exception):
    """Used if the configuration is (partially) invalid."""


# ---------------------------------------------------------------------------
# EXCEPTION/ERROR CLASSES: Related to File Handling
# ---------------------------------------------------------------------------
# -- SINCE: Python 3.3 -- FileNotFoundError is built-in exception
# class FileNotFoundError(LookupError):
#    """Used if a specified file was not found."""
#
class InvalidFileLocationError(LookupError):
    """Used if a :class:`behave.model_core.FileLocation` is invalid.
    This occurs if the file location is no exactly correct and
    strict checking is enabled.
    """


class InvalidFilenameError(ValueError):
    """Used if a filename does not have the expected file extension, etc."""


# ---------------------------------------------------------------------------
# EXCEPTION/ERROR CLASSES: Related to Imported Plugins
# ---------------------------------------------------------------------------
# RELATED: class ModuleNotFoundError(ImportError): -- Since Python 3.6
class ClassNotFoundError(ImportError):
    """Used if module to import exists, but class with this name does not exist."""


class ObjectNotFoundError(ImportError):
    """Used if module to import exists, but object with this name does not exist."""


class InvalidClassError(TypeError):
    """Used if the specified class has the wrong type:

    * not a class
    * not subclass of a required class
    """



# -*- coding: UTF-8 -*-
"""
Supports some active-tags for Python/Python version related functionality.
"""

from __future__ import absolute_import, print_function
import operator
from platform import python_implementation
import sys
import six
from behave.tag_matcher import ValueObject, BoolValueObject


# -----------------------------------------------------------------------------
# CONSTANTS
# -----------------------------------------------------------------------------
PYTHON_VERSION = sys.version_info[:2]
PYTHON_VERSION3 = sys.version_info[:3]


# -----------------------------------------------------------------------------
# HELPERS: ValueObjects
# -----------------------------------------------------------------------------
class VersionValueObject(ValueObject):
    """Provides a ValueObject for version comparisons with version-tuples
    (as: tuple of numbers).
    """

    def __int__(self, value, compare_func=None):
        if isinstance(value, six.string_types):
            value = self.to_version_tuple(value)
        super(VersionValueObject, self).__init__(value, compare_func)

    def matches(self, tag_value):
        try:
            tag_version = self.to_version_tuple(tag_value)
            return super(VersionValueObject, self).matches(tag_version)
        except (TypeError, ValueError) as e:
            self.on_type_conversion_error(tag_value, e)

    @staticmethod
    def to_version_tuple(version):
        if isinstance(version, tuple):
            # -- ASSUME: tuple of numbers
            return version
        elif isinstance(version, six.string_types):
            # -- CONVERT: string-to-tuple of numbers
            return tuple([int(x) for x in version.split(".")])

        # -- OTHERWISE:
        raise TypeError("Expect: string or tuple")


# -----------------------------------------------------------------------------
# ACTIVE-TAGS
# -----------------------------------------------------------------------------
ACTIVE_TAG_VALUE_PROVIDER = {
    "python2": BoolValueObject(six.PY2),
    "python3": BoolValueObject(six.PY3),
    "python.version": "%s.%s" % PYTHON_VERSION,
    "python.min_version": VersionValueObject(PYTHON_VERSION, operator.ge),
    "python.max_version": VersionValueObject(PYTHON_VERSION, operator.le),

    "os": sys.platform.lower(),
    "platform": sys.platform,

    # -- python.implementation: cpython, pypy, jython, ironpython
    "python.implementation": python_implementation().lower(),
    "pypy": BoolValueObject("__pypy__" in sys.modules),
}



# -*- coding: UTF-8 -*-
# HINT: Import adapter only

from cucumber_tag_expressions.model import Expression, Literal, And, Or, Not

# -----------------------------------------------------------------------------
# PATCH TAG-EXPRESSION BASE-CLASS:
# -----------------------------------------------------------------------------
def _Expression_check(self, tags):
    """Checks if tags match this tag-expression.
    NOTE: Backward-compatible to tag-expressions v1.

    :param tags:  Tags (as list of strings)
    :return: True, if tag-expression matches tags.
    :return: False, otherwise.
    """
    return self.evaluate(tags)


Expression.check = _Expression_check
